import logging
import os
from typing import Dict, List, Union, DefaultDict
from collections import defaultdict

import aceutils.file_util as file_util
import aceutils.general_util as gen_util
from acedeploy.core.model_git_entities import GitFile
from acedeploy.core.model_prepostdeploymentsteps import (
    PreOrPostDeploymenStep,
    PreOrPostDeploymentScriptsExecutionOptions,
)
from acedeploy.core.model_solution_entities import (
    SolutionObject,
    SolutionParametersObject,
    SolutionSchema,
)
from acedeploy.core.model_sql_entities import DbObjectType
from aceutils.logger import LoggingAdapter
from acedeploy.core.model_configuration import PreOrPostdeploymentConfigList

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


class SolutionClient(object):
    """
    Object that collects all functionality based around the acedeploy physical solution folder.
    All operations and information based upon these files is stored here.
    """

    def __init__(
        self,
        project_folder: str,
        predeployment_folders: List[str] = [],
        postdeployment_folders: List[str] = [],
        predeployment_settings: PreOrPostdeploymentConfigList = PreOrPostdeploymentConfigList(),
        postdeployment_settings: PreOrPostdeploymentConfigList = PreOrPostdeploymentConfigList(),
        prepostdeployment_filter_list: List[str] = None,
        config_schema_list: Dict[str, List[str]] = None,
        config_project_folder_filter: List[str] = None,
        disabled_object_types: List[DbObjectType] = None,
    ) -> None:
        """
            Init a new SolutionClient
        Args:
            project_folder: str - path of the root folder of the SQL project. This folder contains all the schemas.
            DEPRECATED OPTION - predeployment_folders: List[str] = [] - list of folders that contain predeployment scripts
            DEPRECATED OPTION - postdeployment_folders: List[str] = [] - list of folders that contain postdeployment scripts
            predeployment_settings: PreOrPostdeploymentConfigList - Settings for the predeployments
            postdeployment_settings: PreOrPostdeploymentConfigList - Settings for the postdeployments
            prepostdeployment_filter_list: List[str]=[] - optional, list of regular expressions (case insensitive). If any regex matches against the content, an exception will be raised
            schema_list: Dict[str, List[str]] - optional, dict of schemas, either blacklist or whitelist
            config_project_folder_filter: List[str] - optional, list of expressions to match against filenames. matches are removed.
            disabled_object_types: List[DbObjectType] - optional, list of sql object types which are not allowed in the solution
        """
        if (bool(predeployment_folders) or bool(postdeployment_folders)) and (
            bool(predeployment_settings.items) or bool(postdeployment_settings.items)
        ):
            raise ValueError(
                "You can either supply predeployment_folders and postdeployment_folders or predeployment_settings and postdeployment_settings, but not parameters from both pairs at the same time."
            )

        # legacy support for deprecated arguments
        if bool(predeployment_folders) or bool(postdeployment_folders):
            log.warning(
                "The settings 'preDeployment' and 'postDeployment' are deprecated. Consider using 'preDeploymentSettings' and 'postDeploymentSettings' instead"
            )
            predeployment_folders = (
                [] if predeployment_folders is None else predeployment_folders
            )
            postdeployment_folders = (
                [] if postdeployment_folders is None else postdeployment_folders
            )
            self.predeployment_settings = PreOrPostdeploymentConfigList.from_dict(
                [
                    {"path": f, "type": "folder", "condition": "onChange"}
                    for f in predeployment_folders
                ],
                "",
            )
            self.postdeployment_settings = PreOrPostdeploymentConfigList.from_dict(
                [
                    {"path": f, "type": "folder", "condition": "onChange"}
                    for f in postdeployment_folders
                ],
                "",
            )

        else:
            self.predeployment_settings = predeployment_settings
            self.postdeployment_settings = postdeployment_settings
            predeployment_folders = [
                item.full_path
                for item in self.predeployment_settings.items
                if item.type == "folder"
            ]
            postdeployment_folders = [
                item.full_path
                for item in self.postdeployment_settings.items
                if item.type == "folder"
            ]

        self.pre_and_postdeployment_folders = (
            predeployment_folders + postdeployment_folders
        )

        self.project_folder = project_folder
        self.config_schema_list = config_schema_list
        self.all_objects: List[Union[SolutionObject, SolutionSchema]] = []
        self.all_objects_dict_by_full_name: DefaultDict[
            str, List[Union[SolutionObject, SolutionSchema]]
        ] = None
        self.all_objects_dict_by_id: DefaultDict[
            str, List[Union[SolutionObject, SolutionSchema]]
        ] = None
        self.parameterobjects_dict_by_name_without_parameters: DefaultDict[
            str, List[Union[SolutionObject, SolutionSchema]]
        ] = None
        self.postdeployment_steps: List[PreOrPostDeploymenStep] = []
        self.predeployment_steps: List[PreOrPostDeploymenStep] = []

        self.prepostdeployment_filter_list = (
            []
            if prepostdeployment_filter_list is None
            else prepostdeployment_filter_list
        )
        self.config_project_folder_filter = (
            []
            if config_project_folder_filter is None
            else config_project_folder_filter
        )
        self.disabled_object_types = (
            []
            if disabled_object_types is None
            else disabled_object_types
        )

    def load_solution(
        self,
        git_changed_file_list: List[GitFile] = None,
        string_replace_dict: Dict[str, str] = None,
        pre_and_postdeployment_execution=PreOrPostDeploymentScriptsExecutionOptions.NONE,
    ) -> None:
        """
            Wrapper function for loading all file information about a acedeploy solution.
            Will load schemas, tables, views, functions, procedures, fileformats, predeployment steps, postdeployment steps
        Args:
            git_changed_file_list: List[GitFile] - optional, contains information on git changes to files
            string_replace_dict: Dict[str, str] - optional, contains information about strings which need to be replaced in SQL DDLs
            pre_and_postdeployment_execution: PreOrPostDeploymentScriptsExecutionOptions - optional, determines which pre/postdeployment scripts will be executed
        """
        if git_changed_file_list is None:
            git_changed_file_list = []
        if string_replace_dict is None:
            string_replace_dict = {}
        log.info(f"LOAD files from [ '{self.project_folder}' ]")
        if not os.path.exists(self.project_folder):
            raise EnvironmentError(
                f"Solution path [ '{self.project_folder}' ] is not valid"
            )

        folder_list = [f.path for f in os.scandir(self.project_folder) if f.is_dir()]
        ddl_folder_list = [
            f for f in folder_list if f not in self.pre_and_postdeployment_folders
        ]

        git_changed_file_dict = {
            f.file_name: f.change_type for f in git_changed_file_list
        }

        self._load_objects(ddl_folder_list, git_changed_file_dict, string_replace_dict)
        self._validate_object_ids_unique()
        self._validate_solution_objects_included_in_schema_list()
        self._populate_all_objects_dict_by_full_name()
        self._populate_all_objects_dict_by_id()
        self._populate_parameterobjects_dict_by_name_without_parameters()

        self._load_pre_and_postdeployment(
            self.predeployment_settings,
            self.postdeployment_settings,
            git_changed_file_dict,
            string_replace_dict,
            self.prepostdeployment_filter_list,
            pre_and_postdeployment_execution,
        )

        log_message = (
            f"FOUND files in [ '{self.project_folder}' ]: {gen_util.generate_object_log_summaries(self.all_objects)}, "
            + f"{len(self.predeployment_steps)} predeployment step(s), "
            + f"{len(self.postdeployment_steps)} postdeployment step(s)"
        )
        log.info(log_message)

    def _load_objects(
        self,
        ddl_folder_list: List[str],
        git_changed_file_dict: Dict[str, str] = None,
        string_replace_dict: Dict[str, str] = None,
    ) -> None:
        """
            Load all database objects in given folders. (This does not include schemas and pre/postdeployment scripts.)
        Args:
            ddl_folder_list: List[str] - list of folder names (absolute paths) from which objects should be loaded
            git_changed_file_dict - Dict[str, str]: optional, contains information on git changes to files
            string_replace_dict: Dict[str, str] - optional, contains information about strings which need to be replaced in SQL DDLs
        """
        if git_changed_file_dict is None:
            git_changed_file_dict = {}
        if string_replace_dict is None:
            string_replace_dict = {}
        for ddl_folder in ddl_folder_list:
            filelist = file_util.get_filelist(ddl_folder, [".sql"])
            filelist = file_util.filter_filelist_negative(filelist, self.config_project_folder_filter)
            for filepath in filelist:
                obj = SolutionObject.factory(
                    filepath,
                    git_changed_file_dict.get(filepath, None),
                    string_replace_dict,
                )
                if obj:
                    if obj.object_type in self.disabled_object_types:
                        raise ValueError(f"DISABLED type [ '{obj.object_type}' ] for FILE [ '{filepath}' ]")
                    self.all_objects.append(obj)
                else:
                    log.info(f"EXCLUDE file [ '{filepath}' ] from solution")

    def _load_pre_and_postdeployment(
        self,
        predeployment_settings: PreOrPostdeploymentConfigList,
        postdeployment_settings: PreOrPostdeploymentConfigList,
        git_changed_file_dict: Dict[str, str] = None,
        string_replace_dict: Dict[str, str] = None,
        regex_filter_list: List[str] = None,
        pre_and_postdeployment_execution=PreOrPostDeploymentScriptsExecutionOptions.NONE,
    ) -> None:
        """
            Load pre- and postdeployment steps
        Args:
            predeployment_settings: PreOrPostdeploymentConfigList - list of predeployment settings
            postdeployment_settings: PreOrPostdeploymentConfigList - list of predeployment settings
            git_changed_file_dict - Dict[str, str]: optional, contains information on git changes to files
            string_replace_dict: Dict[str, str] - optional, contains information about strings which need to be replaced in SQL DDLs
            regex_filter_list: List[str]=[] - optional, contains a list of regular expressions (case insensitive). If any regex matches against the content, an exception will be raised
            pre_and_postdeployment_execution: PreOrPostDeploymentScriptsExecutionOptions - optional, determines which pre/postdeployment scripts will be executed
        """
        if git_changed_file_dict is None:
            git_changed_file_dict = {}
        if string_replace_dict is None:
            string_replace_dict = {}
        if regex_filter_list is None:
            regex_filter_list = []
        self.predeployment_steps = SolutionClient._get_pre_or_postdeployment_steps(
            predeployment_settings,
            git_changed_file_dict,
            string_replace_dict,
            regex_filter_list,
            pre_and_postdeployment_execution,
        )
        self.postdeployment_steps = SolutionClient._get_pre_or_postdeployment_steps(
            postdeployment_settings,
            git_changed_file_dict,
            string_replace_dict,
            regex_filter_list,
            pre_and_postdeployment_execution,
        )

    @staticmethod
    def _get_pre_or_postdeployment_steps(
        prepostdeployment_settings: PreOrPostdeploymentConfigList,
        git_changed_file_dict: Dict[str, str] = None,
        string_replace_dict: Dict[str, str] = None,
        regex_filter_list: List[str] = None,
        pre_and_postdeployment_execution=PreOrPostDeploymentScriptsExecutionOptions.NONE,
    ) -> List[PreOrPostDeploymenStep]:
        """
            Return pre- or postdeployment steps for given config.
        Args:
            prepostdeployment_settings: PreOrPostdeploymentConfigList - list of predeployment settings
            git_changed_file_dict - Dict[str, str]: optional, contains information on git changes to files
            string_replace_dict: Dict[str, str] - optional, contains information about strings which need to be replaced in SQL DDLs
            regex_filter_list: List[str]=[] - optional, contains a list of regular expressions (case insensitive). If any regex matches against the content, an exception will be raised
            pre_and_postdeployment_execution: PreOrPostDeploymentScriptsExecutionOptions - optional, determines which pre/postdeployment scripts will be executed
        """
        if git_changed_file_dict is None:
            git_changed_file_dict = {}
        if string_replace_dict is None:
            string_replace_dict = {}
        if regex_filter_list is None:
            regex_filter_list = []

        steps = []
        for pp_setting in prepostdeployment_settings.items:
            if pp_setting.type == "folder":
                all_files = []
                for root, __, files in os.walk(pp_setting.full_path):
                    for name in files:
                        all_files.append(os.path.join(root, name))

                sorted_files = sorted(all_files)
                for step_path in sorted_files:
                    step = PreOrPostDeploymenStep(
                        path=step_path,
                        git_change_type=git_changed_file_dict.get(step_path, None),
                        string_replace_dict=string_replace_dict,
                        regex_filter_list=regex_filter_list,
                    )
                    step.set_target(pp_setting.target)
                    step.set_execute_step(
                        pre_and_postdeployment_execution, pp_setting.condition
                    )
                    log.debug(
                        f"ADD pre/postdeployment step [ '{step.path}' ] with execute_step set to [ '{step.execute_step}' ]"
                    )
                    steps.append(step)
            else:
                raise ValueError("Pre/postdeployment setting type can only be 'folder'")
        return steps

    def _populate_all_objects_dict_by_full_name(self) -> None:
        self.all_objects_dict_by_full_name: DefaultDict[
            str, List[Union[SolutionObject, SolutionSchema]]
        ] = defaultdict(list)
        for obj in self.all_objects:
            self.all_objects_dict_by_full_name[obj.full_name].append(obj)

    def _populate_all_objects_dict_by_id(self) -> None:
        self.all_objects_dict_by_id: DefaultDict[
            str, List[Union[SolutionObject, SolutionSchema]]
        ] = defaultdict(list)
        for obj in self.all_objects:
            self.all_objects_dict_by_id[obj.id].append(obj)

    def _populate_parameterobjects_dict_by_name_without_parameters(self) -> None:
        self.parameterobjects_dict_by_name_without_parameters: DefaultDict[
            str, List[Union[SolutionObject, SolutionSchema]]
        ] = defaultdict(list)
        for obj in self.all_objects:
            if obj.object_type in (DbObjectType.FUNCTION, DbObjectType.PROCEDURE):
                self.parameterobjects_dict_by_name_without_parameters[
                    obj.name_without_params
                ].append(obj)

    def get_object_by_id(self, object_id: str) -> SolutionObject:
        """
            Return the InstanceObject for the given object_id.
        Args:
            id: str
        """
        result = self.all_objects_dict_by_id[object_id.replace('"', "")]
        if len(result) == 0:
            log.debug(f"OBJECT with [ id='{object_id}' ] NOT in FOUND in solution")
            return None
        elif len(result) > 1:
            raise ValueError(
                f"OBJECT with [ id='{object_id}' ] NOT UNIQUE in solution. Found [ '{len(result)}' ]"
            )
        else:
            return result[0]

    def get_object_by_full_name(self, full_name: str, include_schemas: bool = True) -> SolutionObject:
        """
            Return the SolutionObject for the given object full_name
        Args:
            full_name: str
            include_schemas: bool - Optional (default True), include schemas in the result
        """
        result = self.all_objects_dict_by_full_name[full_name.upper().replace('"', "")]
        if not include_schemas:
            result = [r for r in result if r.object_type != DbObjectType.SCHEMA]
        if len(result) == 0:
            log.debug(f"OBJECT with name [ '{full_name}' ] NOT in FOUND in solution")
            return None
        elif len(result) > 1:
            raise ValueError(
                f"OBJECT with name [ '{full_name}' ] NOT UNIQUE in solution. Found [ '{len(result)}' ]"
            )
        else:
            return result[0]

    def get_parameterobject_by_name_without_params(
        self, name_without_params: str
    ) -> SolutionParametersObject:
        """
            Return the SolutionParametersObject function for the given object name_without_params without the function parameters
        Args:
            name_without_params: str - Name of the function or procedure without parameters, i.e. schema.parameters
        """
        result = self.parameterobjects_dict_by_name_without_parameters[
            name_without_params.upper().replace('"', "")
        ]
        if len(result) == 0:
            log.debug(
                f"OBJECT with name [ '{name_without_params}' ] NOT in FOUND in solution"
            )
            return None
        elif len(result) > 1:
            raise ValueError(
                f"OBJECT with name [ '{name_without_params}' ] NOT UNIQUE in solution. Found [ '{len(result)}' ]"
            )
        else:
            return result[0]

    # todo: richtigen type angeben
    def get_object_by_object(self, obj: SolutionObject) -> SolutionObject:
        """
        Return the matching SolutionObject for the given obj.
        Compares the id of the object.
        """
        return self.get_object_by_id(obj.id)

    def _validate_object_ids_unique(self) -> None:
        """
        Validate that each object id in the solution client is unique.
        Raise exception if not.
        """
        seen = set()
        seen_objects = set()
        duplicates = []
        for obj in self.all_objects:
            if obj.id not in seen:
                seen.add(obj.id)
                seen_objects.add(obj)
            else:
                duplicates.append([o for o in seen_objects if o.id == obj.id][0])
                duplicates.append(obj)
        for obj in duplicates:
            log.info(f"DUPLICATE object [ '{obj.id}' ] in FILE [ '{obj.path}' ]")
        if len(duplicates) > 0:
            raise ValueError(
                "DUPLICATE OBJECTS DETECTED. Object ids (type + schema + name) must be unique."
            )

    def _validate_solution_objects_included_in_schema_list(self):
        """
        Validate if the all loaded solution objects are located in schemas allowed by config_schema_list
        """
        if self.config_schema_list:
            objects_in_filtered_schemas = self._get_solution_objects_not_in_schema_list(
                self.all_objects, self.config_schema_list
            )
            if len(objects_in_filtered_schemas) > 0:
                object_names = [o.full_name for o in objects_in_filtered_schemas]
                raise ValueError(
                    f"OBJECTS [ '{object_names}' ] are NOT ALLOWED by setting projectSchemas [ '{self.config_schema_list}' ]"
                )

    @staticmethod
    def _get_solution_objects_not_in_schema_list(
        solution_objects: List[Union[SolutionObject, SolutionSchema]],
        schema_list: Dict[str, List[str]],
    ) -> List[SolutionObject]:
        """
        Return a list of all solution objects, which are not in schemas allowed by schema_list.

        solution_objects: List of solution objects
        schema_list: Dict[str, List[str]]: dict of schemas, either blacklist or whitelist
        """
        if "whitelist" in schema_list:
            schemas_whitelist = [s.upper() for s in schema_list["whitelist"]]
            objects_in_wrong_schema = [
                o for o in solution_objects if o.schema.upper() not in schemas_whitelist
            ]

        elif "blacklist" in schema_list:
            schemas_blacklist = [s.upper() for s in schema_list["blacklist"]]
            objects_in_wrong_schema = [
                o for o in solution_objects if o.schema.upper() in schemas_blacklist
            ]

        else:
            raise EnvironmentError(
                "MALFORMED config value [ 'schema_list' ] (contains neither blacklist nor whitelist)"
            )
        return objects_in_wrong_schema




from abc import abstractmethod
import copy as copy
import logging
import inspect
from typing import List
import aceutils.dict_and_list_util as dict_and_list_util
from acedeploy.core.model_instance_objects import InstanceObject
from aceaccount.core.model_account_object import AccountObject
from aceaccount.core.model_account_object_statement import AccountObjectStatement
from aceaccount.core.model_account_object_instances import AccountObjectInstance, StorageIntegrationInstance, WarehouseInstance, ShareInstance, DatabaseInstance, ExternalVolumeInstance
from aceaccount.core.model_account_object_sql_entities import AccountObjectActionType, AccountObjectType
from aceservices.snowflake_service import SnowClientConfig

from aceutils.logger import LoggingAdapter

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


class AccountObjectAction(AccountObject):
    """
    Collects information about an account object and the action
    that will be applied to it (CREATE, ALTER, DROP).
    This object can create the statement object (AccountObjectStatement)
    that is used to perform the desired action.
    """

    def __init__(
        self,
        name: str,
        object_type: AccountObjectType,
        action_type: AccountObjectActionType,
        current_instance: AccountObjectInstance = None,
        desired_instance: AccountObjectInstance = None,
    ):
        super().__init__(name, object_type)
        self.action_type = action_type
        self.current_instance = current_instance
        self.desired_instance = desired_instance
        if (self.action_type == AccountObjectActionType.DROP) and not (
            (self.current_instance is not None) and (self.desired_instance is None)
        ):
            raise ValueError(
                "If the action type is DROP, a current instance must be given and a desired instance can not be given"
            )

    @staticmethod
    def factory(
        name: str,
        object_type: AccountObjectType,
        action_type: AccountObjectActionType,
        **kwargs,
    ) -> "AccountObjectAction":
        """
        Generate and return a AccountObjectAction for the given parameters.
        """
        mapping = {
            AccountObjectType.STORAGEINTEGRATION: StorageIntegrationAction,
            AccountObjectType.WAREHOUSE: WarehouseAction,
            AccountObjectType.SHARE: ShareAction,
            AccountObjectType.DATABASE: DatabaseAction,
            AccountObjectType.EXTERNALVOLUME: ExternalVolumeAction
        }
        return mapping[object_type](name=name, action_type=action_type, **kwargs)

    @staticmethod
    def factory_from_instance_object(
        instance_object: AccountObjectInstance,
        action_type: AccountObjectActionType,
        **kwargs,
    ) -> "AccountObjectAction":
        """
        Generate and return a AccountObjectAction for the given a AccountObjectInstance and AccountObjectActionType.
        """
        if action_type not in (
            AccountObjectActionType.DROP,
            AccountObjectActionType.ADD,
        ):
            raise ValueError(
                f"Action of type [ '{action_type}' ] can not be generated using this function. Use AccountObjectAction.factory() instead."
            )
        return AccountObjectAction.factory(
            name=instance_object.name,
            object_type=instance_object.object_type,
            action_type=action_type,
            **kwargs,
        )

    @staticmethod
    def factory_from_instance_objects(
        current_instance: AccountObjectInstance,
        desired_instance: AccountObjectInstance,
        action_type: AccountObjectActionType,
        **kwargs,
    ) -> "AccountObjectAction":
        """
        Generate and return a AccountObjectAction for the given a current InstanceObject, a desired InstanceObject and DbActionType.
        """
        if action_type != AccountObjectActionType.ALTER:
            raise ValueError(
                f"Action of type [ '{action_type}' ] can not be generated using this function. Use AccountObjectAction.factory() instead."
            )
        if str(current_instance) != str(desired_instance):
            raise ValueError(
                f"Name or type of current and desired instance do not match ( current: ['{str(current_instance)}'], desired: ['{str(desired_instance)}'] )"
            )
        return AccountObjectAction.factory(
            name=current_instance.name,
            object_type=current_instance.object_type,
            action_type=action_type,
            current_instance=current_instance,
            desired_instance=desired_instance,
            **kwargs,
        )
    
    def _generate_grant_statements_for_new_objects(desired_instance) -> str:
        """
            Function to generate grant statements from account objects definitions for newly created account objects (which are not in the current state).
        """
        object_domain=AccountObjectType.get_object_domain_from_object_type(desired_instance.object_type)

        grant_statements=[]

        for account_object_privilege in desired_instance.grants:

            for role in desired_instance.grants[account_object_privilege]:
    
                grant_statements.append(f'GRANT {account_object_privilege.upper()} ON {object_domain} {desired_instance.full_name} TO ROLE "{role}";')

        return "\n".join(grant_statements)
    
    def _generate_grant_statements(self) -> str:
        """
            Function to generate grant statements from account objects definitions to align the current state of an account objects to the desired state.
        """
        object_domain=AccountObjectType.get_object_domain_from_object_type(self.object_type)

        grant_statements=[]

        for account_object_privilege in self.desired_instance.grants:

            current_grants_of_privilege = self.current_instance.grants.get(account_object_privilege, [])

            for role in self.desired_instance.grants[account_object_privilege]:
    
                if not role in current_grants_of_privilege:

                    grant_statements.append(f'GRANT {account_object_privilege.upper()} ON {object_domain} {self.desired_instance.full_name} TO ROLE "{role}";')

        return "\n".join(grant_statements)

    def _generate_revoke_statements(self) -> str:
        """
            Function to generate revoke statements from account objects definitions to align the current state of an account objects to the desired state.
        """
        object_domain=AccountObjectType.get_object_domain_from_object_type(self.object_type)

        revoke_statements = []

        for account_object_privilege in self.current_instance.grants:

            if not self.desired_instance.grants:
                desired_grants_of_privilege=[]
            else:
                desired_grants_of_privilege = self.desired_instance.grants.get(account_object_privilege, [])

            for role in self.current_instance.grants[account_object_privilege]:
    
                if not role in desired_grants_of_privilege:

                    revoke_statements.append(f'REVOKE {account_object_privilege.upper()} ON {object_domain} {self.desired_instance.full_name} FROM ROLE "{role}";')

        return "\n".join(revoke_statements)
    
    def _generate_settag_statements_for_new_objects(desired_instance) -> str:
        """
            Function to generate set-tag statements from account objects definitions for newly account objects (which are not in the current state).
        """
        object_domain=AccountObjectType.get_object_domain_from_object_type(desired_instance.object_type)

        settag_statement=""

        set_tags=[]
        for tag, tag_value in desired_instance.tags.items():
            if tag_value:
                set_tags.append(f"{tag} = '{tag_value}'")
        set_tags = ", ".join(set_tags)
        if set_tags:
            settag_statement=f"ALTER {object_domain} {desired_instance.full_name} SET TAG {set_tags};"

        return settag_statement
    
    def _generate_settag_statements(self) -> str:
        """
            Function to generate set-tag statements from account objects definitions to align the current state of a an account objects to the desired state.
        """
        object_domain=AccountObjectType.get_object_domain_from_object_type(self.object_type)

        settag_statement=""

        set_tags=[]
        for tag, tag_value in self.desired_instance.tags.items():
            if not hasattr(self.current_instance, "tags") or tag.lower() not in {tag.lower() for tag in self.current_instance.tags} or tag_value != self.current_instance.tags[tag.lower()]:
                set_tags.append(f"{tag} = '{tag_value}'")
        set_tags = ", ".join(set_tags)
        if set_tags:
            settag_statement = f"ALTER {object_domain} {self.desired_instance.full_name} SET TAG {set_tags};"

        return settag_statement

    def _generate_unsettag_statements(self) -> str:
        """
            Function to generate unset-tag statements from account objects definitions to align the current state of a an account objects to the desired state.
        """
        object_domain=AccountObjectType.get_object_domain_from_object_type(self.object_type)

        unsettag_statement=''
        unset_tags=[]
        for tag, tag_value in self.current_instance.tags.items():
            if tag.lower() not in {tag.lower() for tag in self.desired_instance.tags} or not tag_value:
                unset_tags.append(tag)
        unset_tags = ", ".join(unset_tags)
        if unset_tags:
            unsettag_statement = f"ALTER {object_domain} {self.desired_instance.full_name} UNSET TAG {unset_tags};"

        return unsettag_statement

    @abstractmethod
    def _generate_statement(self, **kwargs):
        pass

    def generate_statement_object(self) -> AccountObjectStatement:
        """
            Generate a AccountObjectStatement object, which contains the SQL code required
            to perform the desired action on the db object.
        Returns:
            AccountObjectStatement - containing the SQL statement(s) to achieve the desired result.
        """
        return AccountObjectStatement(
            name=self.name,
            statement=self._generate_statement(),
            object_type=self.object_type,
        )

    @abstractmethod
    def _generate_statement(self, **kwargs):
        pass

    @staticmethod
    def _alter_statement_error_summary(
        updated_instance: InstanceObject,
        desired_instance: InstanceObject,
        properties_to_skip: List[str] = ("database_name",),
    ) -> str:
        """
        Given two instance objects, return a summary of all properties which differ.
        Optionally accepts a list of properties to skip.
        """
        try:
            messages = []
            updated_dict = updated_instance.__dict__
            desired_dict = desired_instance.__dict__
            for key in updated_dict:
                if not key.startswith("_") and key not in properties_to_skip:
                    if updated_dict[key] != desired_dict[key]:
                        messages.append(
                            f"property [ '{key}' ]: current value [ '{updated_dict[key]}' ], desired value [ '{desired_dict[key]}' ]"
                        )
            return "; ".join(messages)
        except:
            return "Could not determine differences."


class StorageIntegrationAction(AccountObjectAction):
    def __init__(
        self,
        name: str,
        action_type: AccountObjectActionType,
        current_instance: StorageIntegrationInstance = None,
        desired_instance: StorageIntegrationInstance = None,

        **_,
    ):
        super().__init__(
            name=name,
            object_type=AccountObjectType.STORAGEINTEGRATION,
            action_type=action_type,
            current_instance=current_instance,
            desired_instance=desired_instance,
        )

    def _generate_statement(self, **_):
        if self.action_type == AccountObjectActionType.ADD:
            return self._generate_create_statement(self.desired_instance)
        elif self.action_type == AccountObjectActionType.ALTER:
            return self._generate_alter_statement(self.current_instance, self.desired_instance)
        elif self.action_type == AccountObjectActionType.DROP:
            return f"DROP STORAGE INTEGRATION {self.full_name};"
        elif self.action_type == AccountObjectActionType.GRANT:
            return self._generate_grant_statements()
        elif self.action_type == AccountObjectActionType.REVOKE:
            return self._generate_revoke_statements()
        elif self.action_type == AccountObjectActionType.SETTAG:
            return self._generate_settag_statements()
        elif self.action_type == AccountObjectActionType.UNSETTAG:
            return self._generate_unsettag_statements()

    def __str__(self):
        return f"StorageIntegrationAction: {self.id}"

    def __repr__(self):
        return f"StorageIntegrationAction: {self.id}"

    @staticmethod
    def _generate_create_statement(desired_instance: StorageIntegrationInstance) -> str:
        """
        Generate create storage integration statements for storage integrations in the desired state but not in the current state.
        Generate grant statements for those storage integrations. Set tags on those storage integrations.
        """
        statements = []

        allowed_locs_str = dict_and_list_util.list_to_string_representation(desired_instance.storage_allowed_locations)
        blocked_locs_str = dict_and_list_util.list_to_string_representation(desired_instance.storage_blocked_locations)

        comment = desired_instance.comment.replace("'", "''")
        comment = comment.replace("\\", "\\\\")

        if desired_instance.storage_provider == "AZURE":
            create_statement = inspect.cleandoc(f"""
                                                CREATE STORAGE INTEGRATION "{desired_instance.name}"
                                                TYPE = EXTERNAL_STAGE
                                                STORAGE_PROVIDER = {desired_instance.storage_provider}
                                                AZURE_TENANT_ID = '{desired_instance.azure_tenant_id}'
                                                ENABLED = {desired_instance.enabled}
                                                STORAGE_ALLOWED_LOCATIONS = ({allowed_locs_str})
                                                COMMENT = '{comment}'
                                                """)
            
            if desired_instance.storage_blocked_locations:
                create_statement += f"""\nSTORAGE_BLOCKED_LOCATIONS = ({blocked_locs_str})"""
        else:
            error_message = f"The storage provider {desired_instance.storage_provider} is not supported for storage integrations."
            raise ValueError(error_message)

        create_statement = f"""{create_statement};"""

        statements.append(create_statement)

        if hasattr(desired_instance, "tags") and desired_instance.tags:
            tag_statements=AccountObjectAction._generate_settag_statements_for_new_objects(desired_instance)
            statements.append(tag_statements)

        if hasattr(desired_instance, "grants") and desired_instance.grants:
            grant_statements=AccountObjectAction._generate_grant_statements_for_new_objects(desired_instance)
            statements.append(grant_statements)

        return "\n".join(statements)

    @staticmethod
    def _generate_alter_statement(
        current_instance: StorageIntegrationInstance,
        desired_instance: StorageIntegrationInstance,
    ) -> str:
        """
        Generate sql statement(s) to change the current state of a storage integration to the desired state.
        """
        statements = []

        updated_instance = copy.deepcopy(
            current_instance
        )  # variable used to track changes for which statements are added

        # enabled
        if updated_instance.enabled != desired_instance.enabled:
            statements.append(
                f"ALTER STORAGE INTEGRATION {current_instance.full_name} SET ENABLED = {desired_instance.enabled};"
            )
            updated_instance.enabled = desired_instance.enabled

        # comment
        if updated_instance.comment != desired_instance.comment:
            if desired_instance.comment is None:
                statements.append(
                    f"ALTER STORAGE INTEGRATION {current_instance.full_name} UNSET COMMENT;"
                )
            else:
                comment = desired_instance.comment.replace("'", "''")
                comment = comment.replace("\\", "\\\\")

                statements.append(
                    f"ALTER STORAGE INTEGRATION {current_instance.full_name} SET COMMENT = '{comment}';"
                )
            updated_instance.comment = desired_instance.comment

        # storage_allowed_locations
        if updated_instance.storage_allowed_locations != desired_instance.storage_allowed_locations:
            if desired_instance.storage_allowed_locations is not None:
                allowed_locs_str = dict_and_list_util.list_to_string_representation(desired_instance.storage_allowed_locations)
                if allowed_locs_str:
                    statements.append(
                            f"ALTER STORAGE INTEGRATION {current_instance.full_name} SET STORAGE_ALLOWED_LOCATIONS = ({allowed_locs_str});"
                        )
                    updated_instance.storage_allowed_locations = desired_instance.storage_allowed_locations
                else:                
                    error_message = f"Requiered parameter STORAGE_ALLOWED_LOCATIONS for storage integration {current_instance.full_name} not defined."
                    raise ValueError(error_message)
            else:
                error_message = f"Requiered parameter STORAGE_ALLOWED_LOCATIONS for storage integration {current_instance.full_name} not defined."
                raise ValueError(error_message)

        # storage_blocked_locations
        if updated_instance.storage_blocked_locations != desired_instance.storage_blocked_locations:
            if desired_instance.storage_blocked_locations is not None:
                blocked_locs_str = dict_and_list_util.list_to_string_representation(desired_instance.storage_blocked_locations)
                if blocked_locs_str:
                    statements.append(
                            f"ALTER STORAGE INTEGRATION {current_instance.full_name} SET STORAGE_BLOCKED_LOCATIONS = ({blocked_locs_str});"
                        )
                    updated_instance.storage_blocked_locations = desired_instance.storage_blocked_locations
                else:
                    statements.append(
                            f"ALTER STORAGE INTEGRATION {current_instance.full_name} UNSET STORAGE_BLOCKED_LOCATIONS;"
                        )
                    updated_instance.storage_blocked_locations = []
            else:
                statements.append(
                            f"ALTER STORAGE INTEGRATION {current_instance.full_name} UNSET STORAGE_BLOCKED_LOCATIONS;"
                        )
                updated_instance.storage_blocked_locations = None

        # azure tenant id -> only if storage_provider = AZURE
        if desired_instance.storage_provider == "AZURE" and updated_instance.azure_tenant_id != desired_instance.azure_tenant_id:
            statements.append(
                f"ALTER STORAGE INTEGRATION {current_instance.full_name} SET AZURE_TENANT_ID = '{desired_instance.azure_tenant_id}';"
            )
            updated_instance.azure_tenant_id = desired_instance.azure_tenant_id

        if updated_instance == desired_instance:
            return "\n".join(statements)
        else:
            property_error_message = AccountObjectAction._alter_statement_error_summary(
                updated_instance, desired_instance, ["tags", "grants"]
            )
            error_message = f"Required ALTER STORAGE INTEGRATION is not supported for {current_instance.full_name} (for {property_error_message}), refer to https://docs.snowflake.com/en/sql-reference/sql/alter-storage-integration.html"
            raise ValueError(error_message)


class WarehouseAction(AccountObjectAction):
    def __init__(
        self,
        name: str,
        action_type: AccountObjectActionType,
        snow_client_config: SnowClientConfig,
        current_instance: WarehouseInstance = None,
        desired_instance: WarehouseInstance = None,
        
        **_,
    ):
        super().__init__(
            name=name,
            object_type=AccountObjectType.WAREHOUSE,
            action_type=action_type,
            current_instance=current_instance,
            desired_instance=desired_instance,
        )
        self.snow_client_config=snow_client_config

    def _generate_statement(self, **_):
        if self.action_type == AccountObjectActionType.ADD:
            return self._generate_create_statement(self.desired_instance, self.snow_client_config)
        elif self.action_type == AccountObjectActionType.ALTER:
            return self._generate_alter_statement(self.current_instance, self.desired_instance)
        elif self.action_type == AccountObjectActionType.DROP:
            return f"DROP WAREHOUSE {self.full_name};"
        elif self.action_type == AccountObjectActionType.GRANT:
            return self._generate_grant_statements()
        elif self.action_type == AccountObjectActionType.REVOKE:
            return self._generate_revoke_statements()
        elif self.action_type == AccountObjectActionType.SETTAG:
            return self._generate_settag_statements()
        elif self.action_type == AccountObjectActionType.UNSETTAG:
            return self._generate_unsettag_statements()

    def __str__(self):
        return f"WarehouseAction: {self.id}"

    def __repr__(self):
        return f"WarehouseAction: {self.id}"

    @staticmethod
    def _generate_create_statement(desired_instance: WarehouseInstance, snow_client_config: SnowClientConfig) -> str:
        """
        Generate create warehouse statements for warehouses in the desired state but not in the current state and generate grant statements for those warehouses.
        Change the current warehouse back to the configured warehouse, since creating a virtual warehouse automatically sets it as the active/current warehouse for the current session.
        Generate grant statements for those warehouses. Set tags on those warehouses.
        """
        statements=[]

        if snow_client_config.warehouse:

            comment = desired_instance.comment.replace("'", "''")
            comment = comment.replace("\\", "\\\\")

            create_statement = inspect.cleandoc(f"""
                                                CREATE WAREHOUSE "{desired_instance.name}"
                                                INITIALLY_SUSPENDED = TRUE
                                                WAREHOUSE_TYPE = '{desired_instance.type}'
                                                WAREHOUSE_SIZE = '{desired_instance.size}'
                                                MAX_CLUSTER_COUNT = {desired_instance.max_cluster_count}
                                                MIN_CLUSTER_COUNT = {desired_instance.min_cluster_count}
                                                SCALING_POLICY = '{desired_instance.scaling_policy}'
                                                AUTO_SUSPEND = {desired_instance.auto_suspend}
                                                AUTO_RESUME = {desired_instance.auto_resume}
                                                COMMENT = '{comment}'
                                                """)
            
            if desired_instance.resource_monitor and not desired_instance.resource_monitor=='null':
                create_statement += f"\nRESOURCE_MONITOR = '{desired_instance.resource_monitor}'"
            if desired_instance.enable_query_acceleration:
                create_statement += f"\nENABLE_QUERY_ACCELERATION = {desired_instance.enable_query_acceleration}"
            if desired_instance.query_acceleration_max_scale_factor:
                create_statement += f"\nQUERY_ACCELERATION_MAX_SCALE_FACTOR = {desired_instance.query_acceleration_max_scale_factor}"
            if desired_instance.max_concurrency_level:
                create_statement += f"\nMAX_CONCURRENCY_LEVEL = {desired_instance.max_concurrency_level}"
            if desired_instance.statement_queued_timeout_in_seconds:
                create_statement += f"\nSTATEMENT_QUEUED_TIMEOUT_IN_SECONDS = {desired_instance.statement_queued_timeout_in_seconds}"
            if desired_instance.statement_timeout_in_seconds:
                create_statement += f"\nSTATEMENT_TIMEOUT_IN_SECONDS = {desired_instance.statement_timeout_in_seconds}"
            if desired_instance.tags:
                statement_tags_ = []
                for tag, tag_value in desired_instance.tags.items():
                    statement_tags_.append(f"{tag} = '{tag_value}'")
                statement_tags_ = ", ".join(statement_tags_)
                statement_tags = f"\nWITH TAG ({statement_tags_})"
                create_statement += statement_tags

            create_statement = f"""{create_statement};\nUSE WAREHOUSE {snow_client_config.warehouse};"""

        else:
            error_message = f"No default warehouse retrieved. Please check Snowflake client configuration."
            raise ValueError(error_message) 
        
        statements.append(create_statement)

        if hasattr(desired_instance, "grants") and desired_instance.grants:
            grant_statements=AccountObjectAction._generate_grant_statements_for_new_objects(desired_instance)
            statements.append(grant_statements)
            
        return "\n".join(statements)

    @staticmethod
    def _generate_alter_statement(
        current_instance: WarehouseInstance,
        desired_instance: WarehouseInstance,
    ) -> str:
        """
        Generate sql statement(s) to change the current state of a warehouse to the desired state.
        """
        statements = []

        updated_instance = copy.deepcopy(
            current_instance
        )  # variable used to track changes for which statements are added

        #warehouse_type
        if updated_instance.type != desired_instance.type:
            statements.append(
                f"ALTER WAREHOUSE {current_instance.full_name} SET WAREHOUSE_TYPE = '{desired_instance.type}';"
            )
            updated_instance.type = desired_instance.type

        #warehouse_size
        if updated_instance.size != desired_instance.size:
            statements.append(
                f"ALTER WAREHOUSE {current_instance.full_name} SET WAREHOUSE_SIZE = '{desired_instance.size}';"
            )
            updated_instance.size = desired_instance.size

        #max_cluster_count
        if updated_instance.max_cluster_count != desired_instance.max_cluster_count:
            statements.append(
                f"ALTER WAREHOUSE {current_instance.full_name} SET MAX_CLUSTER_COUNT = {desired_instance.max_cluster_count};"
            )
            updated_instance.max_cluster_count = desired_instance.max_cluster_count

        #min_cluster_count
        if updated_instance.min_cluster_count != desired_instance.min_cluster_count:
            statements.append(
                f"ALTER WAREHOUSE {current_instance.full_name} SET MIN_CLUSTER_COUNT = {desired_instance.min_cluster_count};"
            )
            updated_instance.min_cluster_count = desired_instance.min_cluster_count

        #scaling_policy
        if updated_instance.scaling_policy != desired_instance.scaling_policy:
            statements.append(
                f"ALTER WAREHOUSE {current_instance.full_name} SET SCALING_POLICY = '{desired_instance.scaling_policy}';"
            )
            updated_instance.scaling_policy = desired_instance.scaling_policy

        #auto_suspend
        if updated_instance.auto_suspend != desired_instance.auto_suspend:
            statements.append(
                f"ALTER WAREHOUSE {current_instance.full_name} SET AUTO_SUSPEND = {desired_instance.auto_suspend};"
            )
            updated_instance.auto_suspend = desired_instance.auto_suspend

        #auto_resume
        if updated_instance.auto_resume != desired_instance.auto_resume:
            statements.append(
                f"ALTER WAREHOUSE {current_instance.full_name} SET AUTO_RESUME = {desired_instance.auto_resume};"
            )
            updated_instance.auto_resume = desired_instance.auto_resume

        #resource_monitor
        if updated_instance.resource_monitor != desired_instance.resource_monitor:
            if desired_instance.resource_monitor and not desired_instance.resource_monitor=='null':
                statements.append(
                    f"ALTER WAREHOUSE {current_instance.full_name} SET RESOURCE_MONITOR = '{desired_instance.resource_monitor}';"
                )
            else:
                statements.append(
                    f"ALTER WAREHOUSE {current_instance.full_name} UNSET RESOURCE_MONITOR;"
                )
            updated_instance.resource_monitor = desired_instance.resource_monitor

        #comment
        if updated_instance.comment != desired_instance.comment:
            if desired_instance.comment is None:
                statements.append(
                    f"ALTER WAREHOUSE {current_instance.full_name} UNSET COMMENT;"
                )
            else:
                comment = desired_instance.comment.replace("'", "''")
                comment = comment.replace("\\", "\\\\")

                statements.append(
                    f"ALTER WAREHOUSE {current_instance.full_name} SET COMMENT = '{comment}';"
                )
            updated_instance.comment = desired_instance.comment

        #enable_query_acceleration
        if updated_instance.enable_query_acceleration != desired_instance.enable_query_acceleration:
            statements.append(
                f"ALTER WAREHOUSE {current_instance.full_name} SET ENABLE_QUERY_ACCELERATION = {desired_instance.enable_query_acceleration};"
            )
            updated_instance.enable_query_acceleration = desired_instance.enable_query_acceleration

        #query_acceleration_max_scale_factor
        if updated_instance.query_acceleration_max_scale_factor != desired_instance.query_acceleration_max_scale_factor:
            statements.append(
                f"ALTER WAREHOUSE {current_instance.full_name} SET QUERY_ACCELERATION_MAX_SCALE_FACTOR = {desired_instance.query_acceleration_max_scale_factor};"
            )
            updated_instance.query_acceleration_max_scale_factor = desired_instance.query_acceleration_max_scale_factor

        #max_concurrency_level
        if updated_instance.max_concurrency_level != desired_instance.max_concurrency_level:
            statements.append(
                f"ALTER WAREHOUSE {current_instance.full_name} SET MAX_CONCURRENCY_LEVEL = {desired_instance.max_concurrency_level};"
            )
            updated_instance.max_concurrency_level = desired_instance.max_concurrency_level

        #statement_queued_timeout_in_seconds
        if updated_instance.statement_queued_timeout_in_seconds != desired_instance.statement_queued_timeout_in_seconds:
            statements.append(
                f"ALTER WAREHOUSE {current_instance.full_name} SET STATEMENT_QUEUED_TIMEOUT_IN_SECONDS = {desired_instance.statement_queued_timeout_in_seconds};"
            )
            updated_instance.statement_queued_timeout_in_seconds = desired_instance.statement_queued_timeout_in_seconds

        #statement_timeout_in_seconds
        if updated_instance.statement_timeout_in_seconds != desired_instance.statement_timeout_in_seconds:
            statements.append(
                f"ALTER WAREHOUSE {current_instance.full_name} SET STATEMENT_TIMEOUT_IN_SECONDS = {desired_instance.statement_timeout_in_seconds};"
            )
            updated_instance.statement_timeout_in_seconds = desired_instance.statement_timeout_in_seconds

        if updated_instance == desired_instance:
            return "\n".join(statements)
        else:
            property_error_message = AccountObjectAction._alter_statement_error_summary(
                updated_instance, desired_instance, ["tags", "grants"]
            )
            error_message = f"Required ALTER WAREHOUSE is not supported for {current_instance.full_name} (for {property_error_message}), refer to https://docs.snowflake.com/en/sql-reference/sql/alter-warehouse.html"
            raise ValueError(error_message)
        
class ShareAction(AccountObjectAction):
    def __init__(
        self,
        name: str,
        action_type: AccountObjectActionType,
        snow_client_config: SnowClientConfig,
        current_instance: ShareInstance = None,
        desired_instance: ShareInstance = None,
        
        **_,
    ):
        super().__init__(
            name=name,
            object_type=AccountObjectType.SHARE,
            action_type=action_type,
            current_instance=current_instance,
            desired_instance=desired_instance,
        )
        self.snow_client_config=snow_client_config

    def _generate_statement(self, **_):
        if self.action_type == AccountObjectActionType.ADD:
            return self._generate_create_statement(self.desired_instance, self.snow_client_config)
        elif self.action_type == AccountObjectActionType.ALTER:
            return self._generate_alter_statement(self.current_instance, self.desired_instance)
        elif self.action_type == AccountObjectActionType.DROP:
            return f"DROP SHARE {self.full_name};"
        elif self.action_type == AccountObjectActionType.GRANT:
            return self._generate_grant_statements()
        elif self.action_type == AccountObjectActionType.REVOKE:
            return self._generate_revoke_statements()
        elif self.action_type == AccountObjectActionType.SETTAG:
            return self._generate_settag_statements()
        elif self.action_type == AccountObjectActionType.UNSETTAG:
            return self._generate_unsettag_statements()

    def __str__(self):
        return f"ShareAction: {self.id}"

    def __repr__(self):
        return f"ShareAction: {self.id}"

    @staticmethod
    def _generate_create_statement(desired_instance: ShareInstance, snow_client_config: SnowClientConfig) -> str:
        """
        Generate "create share" statements for shares in the desired state but not in the current state.
        Grant usage privlege on database to share. Add accounts to share. Set tags on share.
        """
        statements=[]

        comment = desired_instance.comment.replace("'", "''")
        comment = comment.replace("\\", "\\\\")

        create_statement = inspect.cleandoc(f"""
                                            CREATE SHARE "{desired_instance.name}"
                                            COMMENT = '{comment}';
                                            """)
        
        if desired_instance.database_name:
            create_statement += f"\nGRANT USAGE ON DATABASE {desired_instance.database_name} TO SHARE {desired_instance.name};"
        
        if desired_instance.accounts:
            accounts_str = dict_and_list_util.list_to_string_representation(desired_instance.accounts)
            create_statement += f"\nALTER SHARE {desired_instance.name} SET ACCOUNTS= {accounts_str};"
        
        statements.append(create_statement)

        if hasattr(desired_instance, "tags") and desired_instance.tags:
            tag_statements=AccountObjectAction._generate_settag_statements_for_new_objects(desired_instance)
            statements.append(tag_statements)

        return "\n".join(statements)

    @staticmethod
    def _generate_alter_statement(
        current_instance: ShareInstance,
        desired_instance: ShareInstance,
    ) -> str:
        """
        Generate sql statement(s) to change the current state of a share to the desired state.
        """
        statements = []

        updated_instance = copy.deepcopy(
            current_instance
        )  # variable used to track changes for which statements are added


        # database_name
        if updated_instance.database_name.upper() != desired_instance.database_name.upper():
            if updated_instance.database_name:
                statements.append(
                        f"REVOKE USAGE ON DATABASE {updated_instance.database_name} FROM SHARE {current_instance.full_name};"
                    )
            if desired_instance.database_name:
                statements.append(
                        f"GRANT USAGE ON DATABASE {desired_instance.database_name} TO SHARE {current_instance.full_name};"
                    )
            updated_instance.database_name = desired_instance.database_name

        # accounts
        if updated_instance.accounts != desired_instance.accounts:
            if desired_instance.accounts:
                accounts_str = dict_and_list_util.list_to_string_representation(desired_instance.accounts)
                statements.append(
                        f"ALTER SHARE {current_instance.full_name} SET ACCOUNTS= {accounts_str};"
                    )
            else:
                statements.append(
                        f"ALTER SHARE {current_instance.full_name} UNSET ACCOUNTS;"
                    )
            updated_instance.accounts = desired_instance.accounts

        #comment
        if updated_instance.comment != desired_instance.comment:
            if desired_instance.comment is None:
                statements.append(
                    f"ALTER SHARE {current_instance.full_name} UNSET COMMENT;"
                )
            else:
                comment = desired_instance.comment.replace("'", "''")
                comment = comment.replace("\\", "\\\\")

                statements.append(
                    f"ALTER SHARE {current_instance.full_name} SET COMMENT = '{comment}';"
                )
            updated_instance.comment = desired_instance.comment


        if updated_instance == desired_instance:
            return "\n".join(statements)
        else:
            property_error_message = AccountObjectAction._alter_statement_error_summary(
                updated_instance, desired_instance, ["tags"]
            )
            error_message = f"Required ALTER SHARE is not supported for {current_instance.full_name} (for {property_error_message}), refer to https://docs.snowflake.com/en/sql-reference/sql/alter-share.html"
            raise ValueError(error_message)
        
class DatabaseAction(AccountObjectAction):
    def __init__(
        self,
        name: str,
        action_type: AccountObjectActionType,
        snow_client_config: SnowClientConfig,
        current_instance: DatabaseInstance = None,
        desired_instance: DatabaseInstance = None,
        
        **_,
    ):
        super().__init__(
            name=name,
            object_type=AccountObjectType.DATABASE,
            action_type=action_type,
            current_instance=current_instance,
            desired_instance=desired_instance,
        )
        self.snow_client_config=snow_client_config

    def _generate_statement(self, **_):
        if self.action_type == AccountObjectActionType.ADD:
            return self._generate_create_statement(self.desired_instance, self.snow_client_config)
        elif self.action_type == AccountObjectActionType.ALTER:
            return self._generate_alter_statement(self.current_instance, self.desired_instance)
        elif self.action_type == AccountObjectActionType.DROP:
            return f"DROP DATABASE {self.full_name};"
        elif self.action_type == AccountObjectActionType.GRANT:
            return self._generate_grant_statements()
        elif self.action_type == AccountObjectActionType.REVOKE:
            return self._generate_revoke_statements()
        elif self.action_type == AccountObjectActionType.SETTAG:
            return self._generate_settag_statements()
        elif self.action_type == AccountObjectActionType.UNSETTAG:
            return self._generate_unsettag_statements()

    def __str__(self):
        return f"DatabaseAction: {self.id}"

    def __repr__(self):
        return f"DatabaseAction: {self.id}"

    @staticmethod
    def _generate_create_statement(desired_instance: DatabaseInstance, snow_client_config: SnowClientConfig) -> str:
        """
        Generate "create database" statements for databases in the desired state but not in the current state.
        Set tags on database.
        """
        statements=[]

        comment = desired_instance.comment.replace("'", "''")
        comment = comment.replace("\\", "\\\\")

        create_statement = inspect.cleandoc(f"""
                                            CREATE {desired_instance.transient} DATABASE "{desired_instance.name}"
                                            DATA_RETENTION_TIME_IN_DAYS = {desired_instance.data_retention_time_in_days}
                                            MAX_DATA_EXTENSION_TIME_IN_DAYS  = {desired_instance.max_data_extension_time_in_days}
                                            DEFAULT_DDL_COLLATION  = '{desired_instance.default_ddl_collation}'
                                            COMMENT = '{comment}'
                                            """)
        
        if desired_instance.log_level and not desired_instance.log_level.upper()=='OFF':
            create_statement += f"\nLOG_LEVEL = '{desired_instance.log_level}'"

        if desired_instance.suspend_task_after_num_failures and not desired_instance.suspend_task_after_num_failures==0:
            create_statement += f"\nSUSPEND_TASK_AFTER_NUM_FAILURES = {desired_instance.suspend_task_after_num_failures}"

        if desired_instance.trace_level and not desired_instance.trace_level.upper()=='OFF':
            create_statement += f"\nTRACE_LEVEL = '{desired_instance.trace_level}'"

        if desired_instance.user_task_managed_initial_warehouse_size and not desired_instance.user_task_managed_initial_warehouse_size.upper()=='MEDIUM':
            create_statement += f"\nUSER_TASK_MANAGED_INITIAL_WAREHOUSE_SIZE = '{desired_instance.user_task_managed_initial_warehouse_size}'"

        if desired_instance.user_task_timeout_ms and not desired_instance.user_task_timeout_ms==3600000:
            create_statement += f"\nUSER_TASK_TIMEOUT_MS = {desired_instance.user_task_timeout_ms}"

        if desired_instance.tags:
                statement_tags_ =[]
                for tag, tag_value in desired_instance.tags.items():
                    statement_tags_.append(f"{tag} = '{tag_value}'")
                statement_tags_ = ", ".join(statement_tags_)
                statement_tags = f"\nWITH TAG ({statement_tags_})"
                create_statement += statement_tags

        create_statement = f"""{create_statement};"""

        statements.append(create_statement)

        if hasattr(desired_instance, "grants") and desired_instance.grants:
            grant_statements=AccountObjectAction._generate_grant_statements_for_new_objects(desired_instance)
            statements.append(grant_statements)

        return "\n".join(statements)

    @staticmethod
    def _generate_alter_statement(
        current_instance: DatabaseInstance,
        desired_instance: DatabaseInstance,
    ) -> str:
        """
        Generate sql statement(s) to change the current state of a database to the desired state.
        """
        statements = []

        updated_instance = copy.deepcopy(
            current_instance
        )  # variable used to track changes for which statements are added

        #data_retention_time_in_days
        if updated_instance.data_retention_time_in_days != desired_instance.data_retention_time_in_days:
            if desired_instance.data_retention_time_in_days is None:
                statements.append(
                    f"ALTER DATABASE {current_instance.full_name} UNSET DATA_RETENTION_TIME_IN_DAYS;"
                )
            else:
                statements.append(
                    f"ALTER DATABASE {current_instance.full_name} SET DATA_RETENTION_TIME_IN_DAYS = {desired_instance.data_retention_time_in_days};"
                )
            updated_instance.data_retention_time_in_days = desired_instance.data_retention_time_in_days

        #max_data_extension_time_in_days 
        if updated_instance.max_data_extension_time_in_days  != desired_instance.max_data_extension_time_in_days:
            if desired_instance.max_data_extension_time_in_days is None:
                statements.append(
                    f"ALTER DATABASE {current_instance.full_name} UNSET MAX_DATA_EXTENSION_TIME_IN_DAYS;"
                )
            else:
                statements.append(
                    f"ALTER DATABASE {current_instance.full_name} SET MAX_DATA_EXTENSION_TIME_IN_DAYS = {desired_instance.max_data_extension_time_in_days};"
                )
            updated_instance.max_data_extension_time_in_days = desired_instance.max_data_extension_time_in_days

        #default_ddl_collation  
        if updated_instance.default_ddl_collation  != desired_instance.default_ddl_collation:
            if desired_instance.default_ddl_collation is None:
                statements.append(
                    f"ALTER DATABASE {current_instance.full_name} UNSET DEFAULT_DDL_COLLATION;"
                )
            else:
                statements.append(
                    f"ALTER DATABASE {current_instance.full_name} SET DEFAULT_DDL_COLLATION = '{desired_instance.default_ddl_collation}';"
                )
            updated_instance.default_ddl_collation = desired_instance.default_ddl_collation

        #log_level
        if updated_instance.log_level != desired_instance.log_level:
            if desired_instance.log_level is None:
                statements.append(
                    f"ALTER DATABASE {current_instance.full_name} UNSET LOG_LEVEL;"
                )
            else:
                statements.append(
                    f"ALTER DATABASE {current_instance.full_name} SET LOG_LEVEL = '{desired_instance.log_level}';"
                )
            updated_instance.log_level = desired_instance.log_level

        #suspend_task_after_num_failures
        if updated_instance.suspend_task_after_num_failures != desired_instance.suspend_task_after_num_failures:
            if desired_instance.suspend_task_after_num_failures is None:
                statements.append(
                    f"ALTER DATABASE {current_instance.full_name} UNSET SUSPEND_TASK_AFTER_NUM_FAILURES;"
                )
            else:
                statements.append(
                    f"ALTER DATABASE {current_instance.full_name} SET SUSPEND_TASK_AFTER_NUM_FAILURES = {desired_instance.suspend_task_after_num_failures};"
                )
            updated_instance.suspend_task_after_num_failures = desired_instance.suspend_task_after_num_failures

        #trace_level
        if updated_instance.trace_level != desired_instance.trace_level:
            if desired_instance.trace_level is None:
                statements.append(
                    f"ALTER DATABASE {current_instance.full_name} UNSET TRACE_LEVEL;"
                )
            else:
                statements.append(
                    f"ALTER DATABASE {current_instance.full_name} SET TRACE_LEVEL = '{desired_instance.trace_level}';"
                )
            updated_instance.trace_level = desired_instance.trace_level

        #user_task_managed_initial_warehouse_size
        if updated_instance.user_task_managed_initial_warehouse_size != desired_instance.user_task_managed_initial_warehouse_size:
            if desired_instance.user_task_managed_initial_warehouse_size is None:
                statements.append(
                    f"ALTER DATABASE {current_instance.full_name} UNSET USER_TASK_MANAGED_INITIAL_WAREHOUSE_SIZE;"
                )
            else:
                statements.append(
                    f"ALTER DATABASE {current_instance.full_name} SET USER_TASK_MANAGED_INITIAL_WAREHOUSE_SIZE = '{desired_instance.user_task_managed_initial_warehouse_size}';"
                )
            updated_instance.user_task_managed_initial_warehouse_size = desired_instance.user_task_managed_initial_warehouse_size

        #user_task_timeout_ms
        if updated_instance.user_task_timeout_ms != desired_instance.user_task_timeout_ms:
            if desired_instance.user_task_timeout_ms is None:
                statements.append(
                    f"ALTER DATABASE {current_instance.full_name} UNSET USER_TASK_TIMEOUT_MS;"
                )
            else:
                statements.append(
                    f"ALTER DATABASE {current_instance.full_name} SET USER_TASK_TIMEOUT_MS = {desired_instance.user_task_timeout_ms};"
                )
            updated_instance.user_task_timeout_ms = desired_instance.user_task_timeout_ms

        #comment
        if updated_instance.comment != desired_instance.comment:
            if desired_instance.comment is None:
                statements.append(
                    f"ALTER DATABASE {current_instance.full_name} UNSET COMMENT;"
                )
            else:
                comment = desired_instance.comment.replace("'", "''")
                comment = comment.replace("\\", "\\\\")

                statements.append(
                    f"ALTER DATABASE {current_instance.full_name} SET COMMENT = '{comment}';"
                )
            updated_instance.comment = desired_instance.comment


        if updated_instance == desired_instance:
            return "\n".join(statements)
        else:
            property_error_message = AccountObjectAction._alter_statement_error_summary(
                updated_instance, desired_instance, ["tags", "grants"]
            )
            error_message = f"Required ALTER DATABASE is not supported for {current_instance.full_name} (for {property_error_message}), refer to https://docs.snowflake.com/en/sql-reference/sql/alter-database.html"
            raise ValueError(error_message)
        
class ExternalVolumeAction(AccountObjectAction):
    def __init__(
        self,
        name: str,
        action_type: AccountObjectActionType,
        snow_client_config: SnowClientConfig,
        current_instance: ExternalVolumeInstance = None,
        desired_instance: ExternalVolumeInstance = None,
        
        **_,
    ):
        super().__init__(
            name=name,
            object_type=AccountObjectType.EXTERNALVOLUME,
            action_type=action_type,
            current_instance=current_instance,
            desired_instance=desired_instance,
        )
        self.snow_client_config=snow_client_config

    def _generate_statement(self, **_):
        if self.action_type == AccountObjectActionType.ADD:
            return self._generate_create_statement(self.desired_instance)
        elif self.action_type == AccountObjectActionType.ALTER:
            return self._generate_alter_statement(self.current_instance, self.desired_instance)
        elif self.action_type == AccountObjectActionType.DROP:
            if self.current_instance.used_by_iceberg_tables:
                raise ValueError(f'External volume {self.current_instance.full_name} cannot be dropped because it has active iceberg table(s) using it.')
            else:
                return f"DROP EXTERNAL VOLUME {self.full_name};"
        elif self.action_type == AccountObjectActionType.GRANT:
            return self._generate_grant_statements()
        elif self.action_type == AccountObjectActionType.REVOKE:
            return self._generate_revoke_statements()
        elif self.action_type == AccountObjectActionType.SETTAG:
            return self._generate_settag_statements()
        elif self.action_type == AccountObjectActionType.UNSETTAG:
            return self._generate_unsettag_statements()

    def __str__(self):
        return f"ExternalVolumeAction: {self.id}"

    def __repr__(self):
        return f"ExternalVolumeAction: {self.id}"

    @staticmethod
    def _generate_storage_location_string(storage_location_name: str, storage_location: dict) -> str:
        """
        Generate a string defining a storage location which can be used in CREATE EXTERNAL VOLUME and ALTER EXTERNAL VOLUME statements.
        """

        storage_location_params_str = '( '
        storage_location_params_str += f"  \n name = '{storage_location_name}'"
        for param, value in storage_location.items():
            if param == "encryption":
                encryption_str = ''
                for encryption_param, encryption_value in value.items():
                    encryption_str += f" \n {encryption_param} = '{encryption_value}'"
                storage_location_params_str += f' \n {param} = ( {encryption_str} )'
            else:
                storage_location_params_str += f" \n {param} = '{value}'"
        
        storage_location_params_str += ' )'
        return storage_location_params_str

    def _generate_create_statement(self, desired_instance: ExternalVolumeInstance, or_replace: bool = False) -> str:
        """
        Generate "create external volume" statements for external volumes in the desired state but not in the current state.
        Grant usage privlege on database to external volume. Add accounts to external volume. Set tags on external volume.
        """
        statements=[]

        comment = desired_instance.comment.replace("'", "''")
        comment = comment.replace("\\", "\\\\")

        storage_locations_str = "\n ,".join([self._generate_storage_location_string(storage_location_name, storage_location_params) for storage_location_name, storage_location_params in desired_instance.storage_locations.items()])

        if or_replace:
            or_replace_str = 'OR REPLACE'
        else:
            or_replace_str = ''

        create_statement = inspect.cleandoc(f"""
                                            CREATE {or_replace_str} EXTERNAL VOLUME "{desired_instance.name}"
                                            STORAGE_LOCATIONS = ( {storage_locations_str} )
                                            ALLOW_WRITES = {desired_instance.allow_writes}
                                            COMMENT = '{comment}'
                                            """)
        
        create_statement = f"""{create_statement};"""

        statements.append(create_statement)

        if hasattr(desired_instance, "grants") and desired_instance.grants:
            grant_statements=AccountObjectAction._generate_grant_statements_for_new_objects(desired_instance)
            statements.append(grant_statements)

        if hasattr(desired_instance, "tags") and desired_instance.tags:
            raise ValueError(f'Tags are currently not supported for object type {AccountObjectType.EXTERNALVOLUME.value}. Object name: {desired_instance.name}')

        return "\n".join(statements)

    def _generate_alter_statement(
        self,
        current_instance: ExternalVolumeInstance,
        desired_instance: ExternalVolumeInstance,
    ) -> str:
        """
        Generate sql statement(s) to change the current state of a external volume to the desired state.
        """
        statements = []

        updated_instance = copy.deepcopy(
            current_instance
        )  # variable used to track changes for which statements are added


        # storage_locations

        if updated_instance.storage_locations != desired_instance.storage_locations:
            for desired_storage_location_name, desired_storage_location in desired_instance.storage_locations.items():
                
                desired_storage_location_string = self._generate_storage_location_string(desired_storage_location_name, desired_storage_location)

                if desired_storage_location_name not in updated_instance.storage_locations:
                    statements.append(
                            f"ALTER EXTERNAL VOLUME {updated_instance.full_name} ADD STORAGE_LOCATION = {desired_storage_location_string};"
                        )
            
                if desired_storage_location_name in updated_instance.storage_locations and desired_storage_location != updated_instance.storage_locations[desired_storage_location_name]:

                    if len(updated_instance.storage_locations) == 1:
                        # when altering the only existing storage location there is no option for ALTER EXTERNAL VOLUME
                        # -> altering only works with "remove" + "add" 
                        # -> first "remove" means there is no storage location defined which is not allowed 
                        # -> first "add" means adding a storage location with the same name which is also not allowed 
                        # TODO evaluate if it would be better to use a dummy storage location inbetween instead of "CREATE OR REPLACE"
                        if current_instance.used_by_iceberg_tables:
                            raise ValueError(f"The External volume {current_instance.full_name} cannot be replaced because it has active iceberg table(s) using it. Note: When altering the only existing storage location the external volume is re-created.")
                        else:
                            statements.append(self._generate_create_statement(desired_instance, or_replace= True))

                    else:
                        if updated_instance.active_storage_location.lower() == desired_storage_location_name.lower():
                            raise ValueError(f'The active storage location {desired_storage_location_name} cannot be removed on external volume {current_instance.full_name}. Note: When altering an existing storage location the storage location is updated by "remove and add".')
                        else:
                            statements.append(
                                    f"ALTER EXTERNAL VOLUME {updated_instance.full_name} REMOVE STORAGE_LOCATION '{desired_storage_location_name}';"
                                )
                            statements.append(
                                    f"ALTER EXTERNAL VOLUME {updated_instance.full_name} ADD STORAGE_LOCATION = {desired_storage_location_string};"
                                )


            for updated_storage_location_name in updated_instance.storage_locations:

                if updated_storage_location_name not in desired_instance.storage_locations:
                    if updated_instance.active_storage_location.lower() == updated_storage_location_name.lower():
                        raise ValueError(f'The active storage location {updated_storage_location_name} cannot be removed on external volume {current_instance.full_name}.')
                    else:
                        statements.append(
                                f"ALTER EXTERNAL VOLUME {updated_instance.full_name} REMOVE STORAGE_LOCATION '{updated_storage_location_name}';"
                            )

            updated_instance.storage_locations = desired_instance.storage_locations

        # allow_writes
        if updated_instance.allow_writes != desired_instance.allow_writes:

            statements.append(
                    f"ALTER EXTERNAL VOLUME {current_instance.full_name} SET ALLOW_WRITES =  {desired_instance.allow_writes};"
                )

            updated_instance.allow_writes = desired_instance.allow_writes

        #comment
        if updated_instance.comment != desired_instance.comment:
            if desired_instance.comment is None:
                statements.append(
                    f"ALTER EXTERNAL VOLUME {current_instance.full_name} = NULL;"
                )
            else:
                comment = desired_instance.comment.replace("'", "''")
                comment = comment.replace("\\", "\\\\")

                statements.append(
                    f"ALTER EXTERNAL VOLUME {current_instance.full_name} SET COMMENT = '{comment}';"
                )
            updated_instance.comment = desired_instance.comment


        if updated_instance == desired_instance:
            return "\n".join(statements)
        else:
            property_error_message = AccountObjectAction._alter_statement_error_summary(
                updated_instance, desired_instance, ["tags", "grants"]
            )
            error_message = f"Required ALTER EXTERNAL VOLUME is not supported for {current_instance.full_name} (for {property_error_message}), refer to https://docs.snowflake.com/en/sql-reference/sql/alter-external-volume.html"
            raise ValueError(error_message)


import logging
import os
import time
import json
from typing import Dict, List, Union
from aceaccount.core.model_account_object_instances import AccountObjectInstance
from aceaccount.core.model_account_object_sql_entities import AccountObjectType

import aceutils.file_util as file_util
import aceutils.general_util as gen_util
import aceutils.parallelization_util as parallelization_util
from aceservices.snowflake_service import SnowClient
from aceutils.logger import LoggingAdapter

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


class AccountMetadataService:
    """
    Class to query metadata information from snowflake account objects.
    An instance of this class is not scoped to a database.
    """

    def __init__(
        self,
        enabled_object_types: list,
        snow_client: SnowClient,
        dryrun: bool = False
    ):
        """
            Inits a account metadata service.
        Args:
            snow_client: SnowClient - provides connection a snowflake database
        """
        self._enabled_object_types = enabled_object_types
        self._snow_client = snow_client
        self._dryrun = dryrun

        module_root_folder = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "..", "..")
        )
        
        self._sql_files_path = os.path.abspath(
            os.path.join(module_root_folder, "resources", "sql")
        )
        self.all_account_objects = []

        self.account_objects_parameter_representation = {}

    def get_account_objects_subset(self, object_type: AccountObjectType):
        return [account_object for account_object in self.all_account_objects if account_object.object_type == object_type]

    def get_all_account_objects_metadata(self, max_number_of_threads: int=1, get_grants: bool=True, get_tags:  bool=True) -> None:
        """
        Get all metadata information (e.g. on storage integrations) from the account.
        """
        log.info(f"QUERY account object information for object types: {str(self._enabled_object_types)}")

        mapping = {
            AccountObjectType.STORAGEINTEGRATION: self._get_metadata_storageintegrations,
            AccountObjectType.WAREHOUSE: self._get_metadata_warehouses,
            AccountObjectType.SHARE: self._get_metadata_shares,
            AccountObjectType.DATABASE: self._get_metadata_databases,
            AccountObjectType.EXTERNALVOLUME: self._get_metadata_externalvolumes
        }

        for object_type, metadata_function in mapping.items():
            if object_type.value not in self._enabled_object_types:
                log.debug(
                    f"SKIPPING disabled object type: {object_type.value}"
                )
                continue
            metadata_list = metadata_function(max_number_of_threads, get_grants, get_tags)
            self.all_account_objects.extend(
                [AccountObjectInstance.factory(object_type, m, get_grants, get_tags) for m in metadata_list]
            )

        log.info(
            f"FOUND objects in account: {gen_util.generate_account_object_log_summaries(self.all_account_objects)}"
        )

    def _query_metadata(
        self,
        object_type: AccountObjectType,
        template_path: str,
    ) -> List[Dict]:
        """
        Query metadata for a given object type. Return raw result
        """
        log.debug(f"QUERY object information for TYPE [ '{str(object_type)}' ].")
        query = file_util.load(template_path)
        return self._snow_client.execute_query(query)

    def _get_metadata_storageintegrations(self, max_number_of_threads: int=1, get_grants: bool=True, get_tags: bool=True):
        """
        Get all metadata information for storage integrations from the Snowflake account.
        """
        template_path = os.path.join(
            self._sql_files_path, "get_metadata_storageintegrations.sql"
        )
        object_type = AccountObjectType.STORAGEINTEGRATION
        metadata_list_ = self._query_metadata(
            object_type, template_path
        )
        metadata_list_ = self._extend_metadata_storageintegrations(metadata_list_, max_number_of_threads)

        # filter on storage integrations for which the executing Snowflake role has ownership on
        metadata_list = self._filter_storageintegrations_by_ownership(metadata_list_, max_number_of_threads)

        if get_grants:
            metadata_list = self._extend_metadata_with_grants(metadata_list, max_number_of_threads, object_type)
        if get_tags:
            metadata_list = self._extend_metadata_with_tags(metadata_list, max_number_of_threads, object_type)
        return metadata_list
    
    def _filter_storageintegrations_by_ownership(self, metadata_list: List[Dict], max_number_of_threads: int) -> List[Dict]:
        """
        SHOW STORAGE INTEGRATIONS does not have information about the ownership.
        The Storage Integrations are filtered by Storage Integrations for which the executing Snowflake role has ownership on.
        """

        start_time_filter_storageintegrations_by_ownership = time.time()

        if max_number_of_threads<=1:
            # without threading
            metadata_list_filtered_by_ownership=self._get_storageintegrations_filtered_by_ownership(metadata_list)
        else:
            # with threading
            metadata_list_filtered_by_ownership=parallelization_util.execute_func_in_parallel(self._get_storageintegrations_filtered_by_ownership, metadata_list, max_number_of_threads)

        end_time_filter_storageintegrations_by_ownership = time.time()


        log.info(f"============= Execution Time get_all_account_objects_metadata -> _filter_storageintegrations_by_ownership: {round(end_time_filter_storageintegrations_by_ownership - start_time_filter_storageintegrations_by_ownership, 2)} seconds")

        return metadata_list_filtered_by_ownership
        
    
    def _get_storageintegrations_filtered_by_ownership(self, metadata_list: List[Dict]) -> List[Dict]:
        """
        SHOW STORAGE INTEGRATIONS does not have information about the ownership.
        Here, each integration is queried individually in order to retrieve the ownership roles.
        The Storage Integrations are then filtered by Storage Integrations for which the executing Snowflake role has ownership on.
        """
        
        metadata_list_filtered_by_ownership=[]
        for metadata_entry in metadata_list:
             
            ownership_role = ''
        
            query_template = 'SHOW GRANTS ON INTEGRATION "{name}";'

            query = query_template.format(
                name=metadata_entry["name"],
            )

            result = self._snow_client.execute_query(query, use_dict_cursor=True)

            for row in result:
                if row["privilege"] == 'OWNERSHIP':
                    ownership_role = row["grantee_name"]

            if not ownership_role:
                raise ValueError(f'Ownership on Storage Integration {metadata_entry["name"]} could not be retrieved!')
            
            # filter by executing Snowflake role coincides with the ownership role
            if ownership_role != self._snow_client.connection.role:
                continue
            metadata_list_filtered_by_ownership.append(metadata_entry)

        return metadata_list_filtered_by_ownership
    
    def _extend_metadata_storageintegrations(self, metadata_list: List[Dict], max_number_of_threads: int) -> List[Dict]:
        """
        SHOW STORAGE INTEGRATIONS does not return all required metadata.
        Here, each integration is queried individually.
        """
        start_time_extend_metadata_storageintegrations = time.time()

        if max_number_of_threads<=1:
            # without threading
            metadata_list_extend_metadata_storageintegrations=self._get_extended_metadata_storageintegrations(metadata_list)
        else:
            # with threading
            metadata_list_extend_metadata_storageintegrations=parallelization_util.execute_func_in_parallel(self._get_extended_metadata_storageintegrations, metadata_list, max_number_of_threads)

        end_time_extend_metadata_storageintegrations = time.time()
        log.info(f"============= Execution Time get_all_account_objects_metadata -> _extend_metadata_storageintegrations: {round(end_time_extend_metadata_storageintegrations - start_time_extend_metadata_storageintegrations, 2)} seconds")

        return metadata_list_extend_metadata_storageintegrations
    
    def _get_extended_metadata_storageintegrations(self, metadata_list: List[Dict]) -> List[Dict]:
        """
        SHOW STORAGE INTEGRATIONS does not return all required metadata.
        Here, each integration is queried individually.
        """
        query_template = 'DESC STORAGE INTEGRATION "{name}";'
        for metadata_entry in metadata_list:
            query = query_template.format(
                name=metadata_entry["name"],
            )
            result = self._snow_client.execute_query(query, use_dict_cursor=True)
            for row in result:
                if row["property_type"] == 'List' and row["property_value"] == '':
                    metadata_entry.update({row["property"].lower(): []})
                elif row["property_type"] == 'List':
                    metadata_entry.update({row["property"].lower(): row["property_value"].split(",")})
                else:
                    metadata_entry.update({row["property"].lower(): row["property_value"]})
        return metadata_list


    def _get_metadata_warehouses(self, max_number_of_threads: int=1, get_grants: bool=True, get_tags: bool=True):
        """
        Get all metadata information for warehouses from the Snowflake account.
        """

        template_path = os.path.join(
            self._sql_files_path, "get_metadata_warehouses.sql"
        )
        object_type = AccountObjectType.WAREHOUSE
        metadata_list_ = self._query_metadata(
            object_type, template_path
        )

        metadata_list_ = self._extend_metadata_warehouses(metadata_list_, max_number_of_threads)

        metadata_list=[]
        for metadata_object in metadata_list_:
            # filter on warehouses for which the executing Snowflake role has ownership on
            if metadata_object["owner"] != self._snow_client.connection.role:
                continue
            metadata_list.append(metadata_object)

        if get_grants:
            metadata_list = self._extend_metadata_with_grants(metadata_list, max_number_of_threads, object_type)
        if get_tags:
            metadata_list = self._extend_metadata_with_tags(metadata_list, max_number_of_threads, object_type)
        return metadata_list
    
    def _extend_metadata_warehouses(self, metadata_list: List[Dict], max_number_of_threads: int) -> List[Dict]:
        """
        SHOW WAREHOUSES does not return all required metadata.
        Here, each warehouse is queried individually.
        """
        start_time_extend_metadata_warehouses = time.time()

        if max_number_of_threads<=1:
            # without threading
            metadata_list_extended_warehouses=self._get_extended_metadata_warehouses(metadata_list)
        else:
            # with threading
            metadata_list_extended_warehouses=parallelization_util.execute_func_in_parallel(self._get_extended_metadata_warehouses, metadata_list, max_number_of_threads)

        end_time_extend_metadata_warehouses = time.time()
        log.info(f"============= Execution Time get_all_account_objects_metadata -> _extend_metadata_warehouses: {round(end_time_extend_metadata_warehouses - start_time_extend_metadata_warehouses, 2)} seconds")
        
        return metadata_list_extended_warehouses
    
    def _get_extended_metadata_warehouses(self, metadata_list: List[Dict]) -> List[Dict]:
        """
        SHOW WAREHOUSES does not return all required metadata.
        Here, each warehouse is queried individually.
        """
        for metadata_entry in metadata_list:
            query = f'SHOW PARAMETERS IN WAREHOUSE "{metadata_entry["name"]}";'
            result = self._snow_client.execute_query(query, use_dict_cursor=True)
            for row in result:
                if row["key"].lower()=="max_concurrency_level" or  row["key"].lower()=="statement_queued_timeout_in_seconds" or row["key"].lower()=="statement_timeout_in_seconds":
                    row_value=int(row["value"])
                else:
                    row_value=row["value"]
                metadata_entry.update({row["key"].lower(): row_value})

        return metadata_list
    
    def _get_metadata_shares(self, max_number_of_threads: int=1, get_grants: bool=True, get_tags: bool=True):
        """
        Get all metadata information for shares from the Snowflake account. Ignore inbound shares.
        """
        template_path = os.path.join(
            self._sql_files_path, "get_metadata_shares.sql"
        )
        object_type = AccountObjectType.SHARE
        metadata_list_ = self._query_metadata(
            object_type, template_path
        )

        metadata_list=[]
        for metadata_object in metadata_list_:

            # filter on shares for which the executing Snowflake role has ownership on
            if metadata_object["owner"] != self._snow_client.connection.role:
                continue 

            # filter on outbound shares and ignore inbound shares
            if metadata_object["kind"] == "OUTBOUND":

                metadata_object["name"]=metadata_object["name"].split('.')[-1]
                if metadata_object["to"]:
                    metadata_object["accounts"]=list(map(str.strip, metadata_object["to"].split(',')))
                else:
                    metadata_object["accounts"]=[]
                
                metadata_list.append(metadata_object)

        if get_tags:
            metadata_list = self._extend_metadata_with_tags(metadata_list, max_number_of_threads, object_type)

        return metadata_list
    
    def _get_metadata_databases(self, max_number_of_threads: int=1, get_grants: bool=True, get_tags: bool=True):
        """
        Get all metadata information for databases from the Snowflake account.
        """
        template_path = os.path.join(
            self._sql_files_path, "get_metadata_databases.sql"
        )
        object_type = AccountObjectType.DATABASE
        metadata_list_ = self._query_metadata(
            object_type, template_path
        )
        metadata_list_ = self._extend_metadata_databases(metadata_list_, max_number_of_threads)

        metadata_list=[]
        for metadata_object in metadata_list_:

            # filter on databases for which the executing Snowflake role has ownership on
            if metadata_object["owner"] != self._snow_client.connection.role:
                continue 
            #ignore system-defined databases
            if metadata_object["name"]=='SNOWFLAKE' or metadata_object["name"]=='SNOWFLAKE_SAMPLE_DATA':
                continue
            #ignore imported databases
            if not metadata_object["kind"]=='STANDARD':
                continue

            if metadata_object["options"]=='TRANSIENT':
                metadata_object["transient"]='TRANSIENT'
            else:
                metadata_object["transient"]=''

            metadata_list.append(metadata_object)

        if get_grants:
            metadata_list = self._extend_metadata_with_grants(metadata_list, max_number_of_threads, object_type)
        if get_tags:
            metadata_list = self._extend_metadata_with_tags(metadata_list, max_number_of_threads, object_type)

        return metadata_list
    
    def _get_metadata_externalvolumes(self, max_number_of_threads: int=1, get_grants: bool=True, get_tags: bool=True):
        """
        Get all metadata information for externalvolumes from the Snowflake account.
        """
        template_path = os.path.join(
            self._sql_files_path, "get_metadata_externalvolumes.sql"
        )
        object_type = AccountObjectType.EXTERNALVOLUME
        metadata_list_ = self._query_metadata(
            object_type, template_path
        )

        metadata_list_ = self._extend_metadata_externalvolumes(metadata_list_, max_number_of_threads)

        # filter on external volumes for which the executing Snowflake role has ownership on
        metadata_list = self._filter_externalvolumes_by_ownership(metadata_list_, max_number_of_threads)

        if get_grants:
            metadata_list = self._extend_metadata_with_grants(metadata_list, max_number_of_threads, object_type)

        # in case of a dryrun get information about iceberg tables that are using the external volumes -> in order to raise an error in the dryrun when trying to drop or "create or replace" external volumes used by iceberg tables
        if self._dryrun:
            metadata_list = self._extend_metadata_external_volumes_used_by_iceberg_tables(metadata_list)

        return metadata_list

    def _filter_externalvolumes_by_ownership(self, metadata_list: List[Dict], max_number_of_threads: int) -> List[Dict]:
        """
        SHOW EXTERNAL VOLUMES does not have information about the ownership.
        The External Volumes are filtered by External Volumes for which the executing Snowflake role has ownership on.
        """

        start_time_filter_externalvolumes_by_ownership = time.time()

        if max_number_of_threads<=1:
            # without threading
            metadata_list_filtered_by_ownership=self._get_externalvolumes_filtered_by_ownership(metadata_list)
        else:
            # with threading
            metadata_list_filtered_by_ownership=parallelization_util.execute_func_in_parallel(self._get_externalvolumes_filtered_by_ownership, metadata_list, max_number_of_threads)

        end_time_filter_externalvolumes_by_ownership = time.time()


        log.info(f"============= Execution Time get_all_account_objects_metadata -> _filter_externalvolumes_by_ownership: {round(end_time_filter_externalvolumes_by_ownership - start_time_filter_externalvolumes_by_ownership, 2)} seconds")

        return metadata_list_filtered_by_ownership
        
    def _get_externalvolumes_filtered_by_ownership(self, metadata_list: List[Dict]) -> List[Dict]:
        """
        SHOW EXTERNAL VOLUMES does not have information about the ownership.
        Here, each integration is queried individually in order to retrieve the ownership roles.
        The External Volumes are then filtered by External Volumes for which the executing Snowflake role has ownership on.
        """
        
        metadata_list_filtered_by_ownership=[]
        for metadata_entry in metadata_list:
             
            ownership_role = ''
        
            query_template = 'SHOW GRANTS ON EXTERNAL VOLUME "{name}";'

            query = query_template.format(
                name=metadata_entry["name"],
            )

            result = self._snow_client.execute_query(query, use_dict_cursor=True)

            for row in result:
                if row["privilege"] == 'OWNERSHIP':
                    ownership_role = row["grantee_name"]

            if not ownership_role:
                raise ValueError(f'Ownership on External Volume {metadata_entry["name"]} could not be retrieved!')
            
            # filter by executing Snowflake role coincides with the ownership role
            if ownership_role != self._snow_client.connection.role:
                continue
            metadata_list_filtered_by_ownership.append(metadata_entry)

        return metadata_list_filtered_by_ownership
    
    def _extend_metadata_externalvolumes(self, metadata_list: List[Dict], max_number_of_threads: int) -> List[Dict]:
        """
        SHOW EXTERNAL VOLUMES does not return all required metadata.
        Here, each external volume is queried individually.
        """
        start_time_extend_metadata_externalvolumes = time.time()

        if max_number_of_threads<=1:
            # without threading
            metadata_list_extended_externalvolumes=self._get_extended_metadata_externalvolumes(metadata_list)
        else:
            # with threading
            metadata_list_extended_externalvolumes=parallelization_util.execute_func_in_parallel(self._get_extended_metadata_externalvolumes, metadata_list, max_number_of_threads)

        end_time_extend_metadata_externalvolumes = time.time()
        log.info(f"============= Execution Time get_all_account_objects_metadata -> _extend_metadata_externalvolumes: {round(end_time_extend_metadata_externalvolumes - start_time_extend_metadata_externalvolumes, 2)} seconds")
    
        return metadata_list_extended_externalvolumes
    
    def _get_extended_metadata_externalvolumes(self, metadata_list: List[Dict]) -> List[Dict]:
        """
        SHOW EXTERNAL VOLUMES does not return all required metadata.
        Here, each external volume is queried individually.
        """

        for metadata_entry in metadata_list:
            metadata_entry["storage_locations"]={}
            query = f'DESCRIBE EXTERNAL VOLUME "{metadata_entry["name"]}";'
            result = self._snow_client.execute_query(query, use_dict_cursor=True)
            for row in result:
                if row["property"].lower()=="allow_writes":
                    metadata_entry.update({row["property"].lower(): row["property_value"]})
                elif row["property"].lower()=="active":
                    metadata_entry.update({row["property"].lower(): row["property_value"]})
                elif row["property"].lower().startswith("storage_location_"):
                    storage_location_params = json.loads(row["property_value"])
                    storage_location={}
                    for storage_location_param in storage_location_params:
                        if storage_location_param.lower() in AccountObjectType.get_supported_externalvolume_storage_location_params():
                            if storage_location_param.lower() == "encryption_type":
                                encryption = {"type": storage_location_params["ENCRYPTION_TYPE"] }
                                if "ENCRYPTION_KMS_KEY_ID" in storage_location_params:
                                    encryption["kms_key_id"] = storage_location_params["ENCRYPTION_KMS_KEY_ID"]
                                storage_location["encryption"] = encryption
                            else:
                                storage_location[storage_location_param.lower()] = storage_location_params[storage_location_param]

                    metadata_entry["storage_locations"][storage_location["name"]] = storage_location
                    metadata_entry["storage_locations"][storage_location["name"]].pop("name")
                    
        return metadata_list
    
    def _extend_metadata_external_volumes_used_by_iceberg_tables(self, metadata_list: List[Dict]):
        """
        Get information about external volumes for further validation in regards to Iceberg Tables using the external volumes.
        In case of an dryrun, an error will be raised at a later stage when trying to drop or "create or replace" external volumes used by iceberg tables.
        Note: Executing Snowflake Role needs SELECT privilege on Iceberg Tables.
        """    
        start_time_extend_metadata_external_volumes_used_by_iceberg_tables = time.time()

        query_iceberg_tables_metadata = 'SHOW ICEBERG TABLES IN ACCOUNT;'

        result = self._snow_client.execute_query(query_iceberg_tables_metadata, use_dict_cursor=True)

        for metadata_entry in metadata_list:
            metadata_entry["used_by_iceberg_tables"] = False
            for iceberg_table in result:
                if  metadata_entry["name"].upper() == iceberg_table["external_volume_name"].upper():
                    metadata_entry["used_by_iceberg_tables"] = True

        end_time_extend_metadata_external_volumes_used_by_iceberg_tables = time.time()

        log.info(f"============= Execution Time get_all_account_objects_metadata -> _extend_metadata_external_volumes_used_by_iceberg_tables: {round(end_time_extend_metadata_external_volumes_used_by_iceberg_tables - start_time_extend_metadata_external_volumes_used_by_iceberg_tables, 2)} seconds")

        return metadata_list

    def _extend_metadata_databases(self, metadata_list: List[Dict], max_number_of_threads: int) -> List[Dict]:
        """
        SHOW DATABASES does not return all required metadata.
        Here, each database is queried individually.
        """
        start_time_extend_metadata_databases = time.time()

        if max_number_of_threads<=1:
            # without threading
            metadata_list_extended_databases=self._get_extended_metadata_databases(metadata_list)
        else:
            # with threading
            metadata_list_extended_databases=parallelization_util.execute_func_in_parallel(self._get_extended_metadata_databases, metadata_list, max_number_of_threads)

        end_time_extend_metadata_databases = time.time()
        log.info(f"============= Execution Time get_all_account_objects_metadata -> _extend_metadata_databases: {round(end_time_extend_metadata_databases - start_time_extend_metadata_databases, 2)} seconds")
        
        return metadata_list_extended_databases
    
    def _get_extended_metadata_databases(self, metadata_list: List[Dict]) -> List[Dict]:
        """
        SHOW DATABASES does not return all required metadata.
        Here, each database is queried individually.
        """
        for metadata_entry in metadata_list:
            query = f'SHOW PARAMETERS IN DATABASE "{metadata_entry["name"]}";'
            result = self._snow_client.execute_query(query, use_dict_cursor=True)
            for row in result:
                if row["key"].lower()=="data_retention_time_in_days" or  row["key"].lower()=="max_data_extension_time_in_days" or row["key"].lower()=="suspend_task_after_num_failures" or row["key"].lower()=="user_task_timeout_ms":
                    row_value=int(row["value"])
                else:
                    row_value=row["value"]
                metadata_entry.update({row["key"].lower(): row_value})

        return metadata_list

    def _extend_metadata_with_grants(
        self, metadata_list: List[Dict], max_number_of_threads: int, object_type: AccountObjectType
    ) -> List[Dict]:
        """
        Extend the metadata information of the account objects with grants.
        Here, each account object is queried individually -> optionally with parallelization.
        """
        start_time_extend_metadata_with_grants = time.time()

        object_domain = AccountObjectType.get_object_domain_from_object_type(object_type)

        if max_number_of_threads<=1:
            # without threading
            metadata_list_extended_with_grants=self._get_metadata_grants(object_domain, metadata_list)
        else:
            metadata_list_extended_with_grants=parallelization_util.execute_func_in_parallel(self._get_metadata_grants,metadata_list, max_number_of_threads, object_domain)

        end_time_extend_metadata_with_grants = time.time()
        log.info(f"============= Execution Time get_all_account_objects_metadata -> _extend_metadata_with_grants for object_type {object_type}: {round(end_time_extend_metadata_with_grants - start_time_extend_metadata_with_grants, 2)} seconds")

        return metadata_list_extended_with_grants
    
    def _get_metadata_grants(self, object_domain: str, metadata_list: list[Dict])-> list[Dict]:
        """
        Get all grants on a list of account objects and add them to the metadata information.
        """
        for metadata_entry in metadata_list:
            query = f'SHOW GRANTS ON {object_domain} {metadata_entry["name"]};'
            result = self._snow_client.execute_query(query, use_dict_cursor=True)

            grants={}

            for row in result:
                if not row["privilege"] == 'OWNERSHIP' and row["granted_to"] == 'ROLE':
                    grants.setdefault(row["privilege"].lower(),[]).append(row["grantee_name"])
            
            metadata_entry["grants"]=grants

        return metadata_list
    
    def _extend_metadata_with_tags(        
            self, metadata_list: List[Dict], max_number_of_threads: int, object_type: AccountObjectType, 
    ) -> List[Dict]:
        """
        Get all tags on an account object.
        Here, each account object is queried individually.
        """
        start_time_extend_metadata_with_tags = time.time()
        object_domain = AccountObjectType.get_object_domain_from_object_type(object_type)

        if max_number_of_threads<=1:
            # without threading
            metadata_list_extended_with_tags=self._get_metadata_tags(object_domain, metadata_list)
        else:
            # with threading
            metadata_list_extended_with_tags=parallelization_util.execute_func_in_parallel(self._get_metadata_tags, metadata_list, max_number_of_threads, object_domain)

        end_time_extend_metadata_with_tags = time.time()
        log.info(f"============= Execution Time get_all_account_objects_metadata -> _extend_metadata_with_tags for object_type {object_type}: {round(end_time_extend_metadata_with_tags - start_time_extend_metadata_with_tags, 2)} seconds")

        return metadata_list_extended_with_tags
    
    def _get_metadata_tags(self, object_domain: str, metadata_list: list[Dict])-> list[Dict]:
        """
        Get all tags on a list of account objects and add them to the metadata information.
        """
        for metadata_entry in metadata_list:
            metadata_entry_name=metadata_entry["name"]
            query = f"SELECT * FROM TABLE(SNOWFLAKE.INFORMATION_SCHEMA.TAG_REFERENCES('{metadata_entry_name}', '{object_domain}'));"
            result = self._snow_client.execute_query(query, use_dict_cursor=True)

            tags={}

            for row in result:
                tags[f'{row["TAG_DATABASE"].lower()}.{row["TAG_SCHEMA"].lower()}.{row["TAG_NAME"].lower()}'] = row["TAG_VALUE"]
            
            metadata_entry["tags"]=tags

        return metadata_list
          
    def get_object_by_object(self, obj: AccountObjectInstance) -> AccountObjectInstance:
        """
        Return the matching InstanceObject for the given obj.
        Compares the id of the object.
        """
        return self.get_object_by_id(obj.id)

    def get_object_by_id(self, object_id: str) -> AccountObjectInstance:
        """
        Return the AccountObjectInstance for the given object_id.
        """
        result = [o for o in self.all_account_objects if (o.compare_id(object_id))]
        if len(result) == 0:
            log.debug(
                f"OBJECT with [ id='{object_id}' ] NOT in FOUND in AccountMetadataService"
            )
            return None
        elif len(result) > 1:
            raise ValueError(
                f"OBJECT with [ id='{object_id}' ] NOT UNIQUE in AccountMetadataService. Found [ '{len(result)}' ]"
            )
        else:
            return result[0]

    def get_account_objects_dict(self, json_schema_relative_path:str) -> dict:
        """
        Get parameter representation as a dict of account objects instances for all object types.
        """

        for object_type in AccountObjectType:
            parameter_representation, object_type_key = self._get_parameter_representation(object_type)
            self.account_objects_parameter_representation[object_type]={
                "$schema":json_schema_relative_path,
                 object_type_key: parameter_representation
            }
        return self.account_objects_parameter_representation

    def _get_parameter_representation(self, object_type: AccountObjectType):
        """
        Return parameter representation of account objects instances for a specific object type.
        """
        mapping_object_type_object_type_key = {
            AccountObjectType.STORAGEINTEGRATION: "storage_integrations",
            AccountObjectType.WAREHOUSE: "warehouses",
            AccountObjectType.SHARE: "shares",
            AccountObjectType.DATABASE: "databases",
            AccountObjectType.EXTERNALVOLUME: "external_volumes",
        }

        object_type_key=mapping_object_type_object_type_key[object_type]

        object_instances=self.get_account_objects_subset(object_type)
        
        parameter_representation={}
        for object_instance in object_instances:

            parameter_representation_={
                object_instance.name:{
                    key:value for key, value in object_instance.__dict__.items() if not key.startswith('__') and not callable(key) and not key=="_object_type" and not key=="_name"
                }                                           
            }

            parameter_representation= parameter_representation | parameter_representation_

        return parameter_representation, object_type_key


from typing import List

from acedeploy.core.model_database_object import DatabaseObject
from acedeploy.core.model_sql_entities import DbObjectType


class DbStatement(DatabaseObject):
    """
    Models atomic sql statement execution
    With target object name and sql statement to execute
    """

    def __init__(
        self, name: str, schema: str, statement: str, object_type: DbObjectType
    ):
        super().__init__(schema=schema, name=name, object_type=object_type)
        self.statement = statement

    def __eq__(self, other):
        return (
            self.schema == other.schema
            and self.name == other.name
            and self.statement == other.statement
            and self.object_type == other.object_type
        )

    def __str__(self):
        return f"DbStatement: {self.id}"

    def __repr__(self):
        return f"DbStatement: {self.id}"


class ParametersObjectStatement(DbStatement):
    """
    DbStatements for objects with parameters
    """

    def __init__(
        self,
        name: str,
        schema: str,
        statement: str,
        object_type: DbObjectType,
        parameters: List[str],
    ):
        super().__init__(
            name=name, schema=schema, statement=statement, object_type=object_type
        )
        self.parameters = parameters

    @property
    def id(self):
        return f"{super().id} {self.parameters_string.upper()}"

    @property
    def full_name(self):
        return f"{super().full_name} {self.parameters_string.upper()}"

    @property
    def parameters_string(self):
        return f"({', '.join([s for s in self.parameters])})"

    def __str__(self):
        return f"ParametersObjectStatement: {self.id}"

    def __repr__(self):
        return f"ParametersObjectStatement: {self.id}"



import json
import logging
from datetime import datetime

import aceutils.file_util as file_util
from aceservices.snowflake_service import SnowClient
from aceutils.logger import LoggingAdapter

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


class Smoketest:
    """
    Class to perform a smoketest
    """

    def __init__(
        self,
        snow_client: SnowClient,
        smoketest_config_file: str,
        results_file: str = "",
    ):
        """
            Inits a metadata service.
        Args:
            snow_client: SnowClient - provides connection a snowflake database
            smoketest_config_file: str - path to file which contains the config json
            results_file: str - optional: file in which to store smoketest results
        """
        self._snow_client = snow_client
        if not self._validate_json(smoketest_config_file):
            raise ValueError("The given smoktest config is not valid")
        self._config = file_util.load_json(smoketest_config_file)
        self._results_file = results_file
        self.results_succeeded = []
        self.results_failed = []

    def _validate_json(self, smoketest_config_file: str):
        """
        Validate if the given json string contains a valid config
        """
        json_schema_file = file_util.get_path(
            ["resources", "json-schemas", "smoketest.schema.json"]
        )
        return file_util.validate_json(json_schema_file, smoketest_config_file)

    def start_smoketest(self):
        """
        Start the smoketest execution
        """
        log.info("START SMOKETEST")

        self.results_succeeded = []
        self.results_failed = []

        s, f = self.smoketest_explain_select_from_views()
        self.results_succeeded.extend(s)
        self.results_failed.extend(f)

        s, f = self.smoketest_explain_select_from_tables()
        self.results_succeeded.extend(s)
        self.results_failed.extend(f)

        log.info(
            f"FINISH SMOKETEST. Ran [ '{len(self.results_succeeded)+len(self.results_failed)}' ] tests. Success: [ '{len(self.results_succeeded)}' ]. Failed: [ '{len(self.results_failed)}' ]."
        )
        if len(self.results_succeeded) > 0:
            log.info(
                f"List of successful tests:\n{json.dumps(self.results_succeeded, indent=4)}"
            )
        if len(self.results_failed) > 0:
            log.info(
                f"List of failed tests:\n{json.dumps(self.results_failed, indent=4)}"
            )

        if self._results_file != "":
            self.save_results()

    def smoketest_explain_select_from_views(self):
        """
        Run a select for all views.
        """
        log.info(
            f"RUN test [ 'smoketest_explain_select_from_views' ] on DATABASE [ '{self._config['database']}' ] with ROLE [ '{self._config['role']}' ]"
        )
        succeeded, failed = self.explain_select_from_object(
            self._config["objects"]["views"],
            "view",
            "smoketest_explain_select_from_views",
        )
        log.info("FINISH test [ 'smoketest_explain_select_from_views' ]")
        return succeeded, failed

    def smoketest_explain_select_from_tables(self):
        """
        Run a select for all tables.
        """
        log.info(
            f"RUN test [ 'smoketest_explain_select_from_tables' ] on DATABASE [ '{self._config['database']}' ] with ROLE [ '{self._config['role']}' ]"
        )
        succeeded, failed = self.explain_select_from_object(
            self._config["objects"]["tables"],
            "table",
            "smoketest_explain_select_from_tables",
        )
        log.info("FINISH test [ 'smoketest_explain_select_from_tables' ]")
        return succeeded, failed

    def explain_select_from_object(self, objectnames, objecttype, testname):
        """
        Run a select for all given objects.
        """
        succeeded = []
        failed = []
        self._snow_client.execute_statement(f"USE ROLE {self._config['role']};")
        for obj in objectnames:
            query = f"EXPLAIN SELECT * FROM {self._config['database']}.{obj};"
            try:
                self._snow_client.execute_query(query)
                succeeded.append(
                    {
                        "name": obj,
                        "type": objecttype,
                        "test": testname,
                        "query": query,
                        "timestamp": datetime.utcnow().isoformat(),
                    }
                )
            except Exception as e:
                failed.append(
                    {
                        "name": obj,
                        "type": objecttype,
                        "test": testname,
                        "query": query,
                        "timestamp": datetime.utcnow().isoformat(),
                        "error": str(e),
                    }
                )
        return succeeded, failed

    def save_results(self):
        """
        Save all results in a JSON file
        """
        log.info(f"SAVE results as [ '{self._results_file}' ]")
        content = {
            "config": self._config,
            "results": {
                "failed": self.results_failed,
                "succeeded": self.results_succeeded,
            },
        }
        file_util.save_json(self._results_file, content)
